from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

from lite_dist2.common import hex2int
from lite_dist2.expections import LD2InvalidSpaceError, LD2ParameterError, LD2TypeError, LD2UndefinedError
from lite_dist2.interfaces import Mergeable
from lite_dist2.value_models.base_space import FlattenSegment, ParameterSpace
from lite_dist2.value_models.line_segment import (
    DummyLineSegment,
    LineSegment,
    LineSegmentModel,
    ParameterRangeBool,
    ParameterRangeFloat,
    ParameterRangeInt,
)
from lite_dist2.value_models.parameter_aligned_space_helper import infinite_product
from lite_dist2.value_models.point import ParamType, ScalarValue

if TYPE_CHECKING:
    from collections.abc import Generator

    from lite_dist2.type_definitions import PrimitiveValueType


class ParameterAlignedSpaceModel(BaseModel):
    type: Literal["aligned"]
    axes: list[LineSegmentModel]
    check_lower_filling: bool

    def get_total(self) -> int | None:
        t = 1
        for axis in self.axes:
            if axis.size is None:
                return None
            t *= hex2int(axis.size)
        return t


class ParameterAlignedSpace(ParameterSpace, Mergeable):
    def __init__(self, axes: list[LineSegment], check_lower_filling: bool) -> None:
        self.axes = axes  # larger index, deeper dimension
        self.filling_dim = [axis.is_universal() for axis in self.axes]
        self.check_lower_filling = check_lower_filling

        if self.check_lower_filling:
            self._check_lower_filling()

    def __hash__(self) -> int:
        # for cache
        return hash(tuple(hash(axis) for axis in self.axes))

    def __eq__(self, other: object) -> bool:
        # for cache and test
        if isinstance(other, ParameterAlignedSpace):
            return self.axes == other.axes
        return False

    def _check_lower_filling(self) -> None:
        #  F  F  F  T  T
        #  1  1  3  10 10
        #  i  f  f  f  f
        # のような場合のみ許可する
        for axis in self.axes[1:]:
            if axis.ambient_size is None:
                msg = "Infinite dimension is only allowed at the first dimension"
                raise LD2InvalidSpaceError(msg)

        is_lower_filled = False
        min_filled_dim = -1
        for i, fill in enumerate(self.filling_dim):
            if not is_lower_filled and fill:
                min_filled_dim = i
                is_lower_filled = True
                continue
            if is_lower_filled and (not fill):
                msg = "Filling from lower dimension"
                raise LD2InvalidSpaceError(msg)

        for i, axis in enumerate(self.axes):
            if i == self.dim - 1:
                continue
            if min_filled_dim == -1 and axis.size != 1:
                # No axis is universal
                msg = "Upper dimension must be size=1"
                raise LD2InvalidSpaceError(msg)
            if i < min_filled_dim - 1 and axis.size != 1:
                msg = "Upper dimension must be size=1"
                raise LD2InvalidSpaceError(msg)

    @functools.cached_property
    def dim(self) -> int:
        return self.get_dim()

    def get_dim(self) -> int:
        return len(self.dimensional_sizes)

    @functools.cached_property
    def dimensional_sizes(self) -> tuple[int, ...]:
        return self.get_dimensional_sizes()

    def get_dimensional_sizes(self) -> tuple[int, ...]:
        return tuple(axis.size for axis in self.axes)

    @functools.cached_property
    def total(self) -> int | None:
        return self.get_total()

    def get_total(self) -> int | None:
        t = 1
        for axis in self.axes:
            if axis.size is None:
                return None
            t *= axis.size
        return t

    def lower_element_num_by_dim(self) -> tuple[int, ...]:
        return self.get_lower_element_num_by_dim([axis.ambient_size for axis in self.axes])

    def is_infinite(self) -> bool:
        return self.total is None

    @functools.cached_property
    def dummy_info(self) -> list[DummyLineSegment]:
        return self.get_dummy_info()

    def get_dummy_info(self) -> list[DummyLineSegment]:
        return [axis.to_dummy() for axis in self.axes]

    def grid(self) -> Generator[tuple[PrimitiveValueType, ...], None, None]:
        yield from infinite_product(*(axis.grid() for axis in self.axes))

    def get_flatten_ambient_start_and_size(self) -> FlattenSegment:
        if not self.check_lower_filling:
            msg = "Cannot get flatten info. Because check_lower_filling of this space is False."
            raise LD2InvalidSpaceError(msg)

        lower_element_num_by_dim = self.lower_element_num_by_dim()
        flatten_index = 0
        for di in range(self.dim):
            flatten_index += self.axes[di].ambient_index * lower_element_num_by_dim[di]
        return FlattenSegment(flatten_index, self.total)

    def get_flatten_ambient_start_and_size_list(self) -> list[FlattenSegment]:
        return [self.get_flatten_ambient_start_and_size()]

    def get_lower_not_universal_dim(self) -> int:
        for i in reversed(range(self.dim)):
            if not self.axes[i].is_universal():
                return i
        return -1

    def indexed_grid(self) -> Generator[tuple[tuple[int, PrimitiveValueType], ...], None, None]:
        yield from infinite_product(*(axis.indexed_grid() for axis in self.axes))

    def slice(self, start_and_sizes: list[tuple[int, int]]) -> ParameterAlignedSpace:
        if len(start_and_sizes) != self.dim:
            msg = "start_and_sizes"
            raise LD2ParameterError(msg, "different size to axes")

        axes = [self.axes[i].slice(*start_and_sizes[i]) for i in range(self.dim)]
        return ParameterAlignedSpace(axes=axes, check_lower_filling=self.check_lower_filling)

    def value_tuple_to_param_type(self, values: tuple[PrimitiveValueType, ...]) -> ParamType:
        # TODO: type="vector" にも対応させる
        return tuple(
            [
                ScalarValue.create_from_numeric(val, ax.type, ax.name)
                for val, ax in zip(values, self.axes, strict=True)
            ],
        )

    def derived_by_same_ambient_space_with(self, other: ParameterSpace) -> bool:
        if not isinstance(other, ParameterAlignedSpace):
            return False
        if len(self.axes) != len(other.axes):
            return False
        return all(s.derived_by_same_ambient_space_with(o) for s, o in zip(self.axes, other.axes, strict=True))

    def get_start_index(self, *args: object) -> int:
        target_dim = self._check_arg(*args)
        return self.axes[target_dim].get_start_index()

    def can_merge(self, other: ParameterAlignedSpace, *args: object) -> bool:
        target_dim = self._check_arg(*args)

        if not self.derived_by_same_ambient_space_with(other):
            # 同じ母空間から誘導されたものでなければ False
            return False

        if self.filling_dim != other.filling_dim:
            # 各次元の占有状態が同じでなければ False
            return False

        if self.filling_dim[target_dim]:
            # 対象の次元を占有していたら False
            return False
        if not all(self.filling_dim[target_dim + 1 :]):
            # 対象の次元より深い次元を占有していなければ False
            # 最深次元(空リスト)であれば True
            return False

        if not all(self.axes[d] == other.axes[d] for d in range(target_dim)):
            # target_dim より浅層の各次元が一致していなければ False
            return False

        self_axis = self.axes[target_dim]
        other_axis = other.axes[target_dim]
        return self_axis.can_merge(other_axis)

    def merge(self, other: ParameterAlignedSpace, *args: object) -> ParameterAlignedSpace:
        target_dim = self._check_arg(*args)

        axes = []
        for d in range(self.dim):
            if d != target_dim:
                # 浅層次元か深層次元: 同じはずなのでそのままコピー
                axes.append(self.axes[d])
                continue

            # target_dim
            merged_axis = self.axes[d].merge(other.axes[d])
            axes.append(merged_axis)

        return ParameterAlignedSpace(axes=axes, check_lower_filling=self.check_lower_filling)

    def to_aligned_list(self) -> list[ParameterAlignedSpace]:
        return [self]

    def to_model(self) -> ParameterAlignedSpaceModel:
        return ParameterAlignedSpaceModel(
            type="aligned",
            axes=[axis.to_model() for axis in self.axes],
            check_lower_filling=self.check_lower_filling,
        )

    @staticmethod
    def loom_by_flatten_index(flatten_index: int, lower_element_num_by_dim: tuple[int, ...]) -> tuple[int, ...]:
        # x -> (a_i, b_j, c_k)
        residual_flatten_index = flatten_index
        loomed_indices = []
        for size in lower_element_num_by_dim:
            loomed_indices.append(residual_flatten_index // size)
            residual_flatten_index = residual_flatten_index % size
        return tuple(loomed_indices)

    @staticmethod
    def _check_arg(*args: object) -> int:
        if len(args) < 1:
            name = "target_dim"
            raise LD2ParameterError(name, "missing")

        target_dim = args[0]
        if not isinstance(target_dim, int):
            name = "target_dim"
            raise LD2TypeError(name, int, type(args[0]))
        return target_dim

    @staticmethod
    def from_model(space_model: ParameterAlignedSpaceModel) -> ParameterAlignedSpace:
        axes = []
        for axis_model in space_model.axes:
            if axis_model.is_dummy:
                param = f"{LineSegmentModel.__name__}.type"
                msg = f"An axis of {ParameterAlignedSpace.__name__} is not allowed dummy axis."
                raise LD2ParameterError(param, msg)

            match axis_model.type:
                case "bool":
                    axis = ParameterRangeBool.from_model(axis_model)
                case "int":
                    axis = ParameterRangeInt.from_model(axis_model)
                case "float":
                    axis = ParameterRangeFloat.from_model(axis_model)
                case _:
                    raise LD2UndefinedError(axis_model.type)
            axes.append(axis)

        return ParameterAlignedSpace(axes=axes, check_lower_filling=space_model.check_lower_filling)



import binascii
import hashlib
import time
from pathlib import Path
from typing import Any

from lite_dist2.config import TableConfig, TableConfigProvider, WorkerConfig
from lite_dist2.curriculum_models.study_portables import StudyRegistry
from lite_dist2.study_strategies import StudyStrategyModel
from lite_dist2.study_strategies.base_study_strategy import StudyStrategyParam
from lite_dist2.suggest_strategies import SuggestStrategyModel
from lite_dist2.suggest_strategies.base_suggest_strategy import SuggestStrategyParam
from lite_dist2.table_node_api.start_table_api import start_in_thread
from lite_dist2.table_node_api.table_param import StudyRegisterParam
from lite_dist2.type_definitions import RawParamType, RawResultType
from lite_dist2.value_models.aligned_space_registry import LineSegmentRegistry, ParameterAlignedSpaceRegistry
from lite_dist2.value_models.point import ScalarValue
from lite_dist2.worker_node.table_node_client import TableNodeClient
from lite_dist2.worker_node.trial_runner import AutoMPTrialRunner
from lite_dist2.worker_node.worker import Worker


def set_table_config() -> None:
    config = TableConfig(
        port=8000,
        curriculum_path=Path(__file__).parent / "example_curriculum.json",
        curriculum_save_interval_seconds=10,
    )
    TableConfigProvider.set(config)


def register_study(table_ip: str, table_port: int) -> None:
    _resolution = 10
    _half_size = 2.0

    study_register_param = StudyRegisterParam(
        study=StudyRegistry(
            name="preimage",
            required_capacity=set(),
            study_strategy=StudyStrategyModel(
                type="find_exact",
                study_strategy_param=StudyStrategyParam(
                    target_value=ScalarValue(
                        type="scalar",
                        value_type="int",
                        value="0x49f68a5c8493ec2c0bf489821c21fc3b",
                    ),
                ),
            ),
            suggest_strategy=SuggestStrategyModel(
                type="sequential",
                suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
            ),
            result_type="scalar",
            result_value_type="int",
            parameter_space=ParameterAlignedSpaceRegistry(
                type="aligned",
                axes=[
                    LineSegmentRegistry(
                        name="x",
                        type="int",
                        size=None,
                        step="0x1",
                        start="0x0",
                    ),
                ],
            ),
        ),
    )
    client = TableNodeClient(table_ip, table_port)
    client.register_study(study_register_param)


class MD5Preimage(AutoMPTrialRunner):
    def func(self, parameters: RawParamType, *_: tuple[Any, ...], **__: dict[str, Any]) -> RawResultType:
        preimage = int(parameters[0])
        hex_digest = hashlib.md5(self.to_bytes(preimage)).hexdigest()  # noqa: S324
        return self.from_hex(hex_digest)

    @staticmethod
    def to_bytes(val: int) -> bytes:
        hex_str = hex(val)[2:]
        if len(hex_str) % 2 == 1:
            hex_str = "0" + hex_str
        return binascii.unhexlify(hex_str)

    @staticmethod
    def from_hex(hex_str: str) -> int:
        return int("0x" + hex_str, base=16)


def run_worker(table_ip: str, table_port: int) -> None:
    worker_config = WorkerConfig(
        name="w_01",
        process_num=2,
        max_size=2000,
        wait_seconds_on_no_trial=5,
        table_node_request_timeout_seconds=60,
    )
    worker = Worker(
        trial_runner=MD5Preimage(),
        ip=table_ip,
        port=table_port,
        config=worker_config,
    )
    worker.start(stop_at_no_trial=True)


if __name__ == "__main__":
    # 1. Set table config for example
    #    Execution node type in normal use: Table
    #    NOTE: For normal use, change the `table_config.json` file in the root directory.
    set_table_config()

    # 2. Start table node in other thread
    #    Execution node type in normal use: Table
    #    NOTE: In most cases, use `start()` (or `uv run start`) instead of `start_in_thread()`
    #          since table nodes and worker nodes are separated.
    table_thread = start_in_thread()
    time.sleep(1)  # wait for activate

    # 3. Register study to table node
    #    Execution node type in normal use: Management
    #    NOTE: If the management node is also a worker node or a table node,
    #          `TableNodeClient.register_study` is available.
    #          Otherwise, studies can be registered using the curl command.
    register_study(table_ip="127.0.0.1", table_port=8000)

    # 4. run worker node
    #    Execution node type in normal use: Worker
    #    NOTE: In this example, the entire process should be completed within 10 seconds.
    #          10 seconds later, example_curriculum.json should be saved in the same directory as this file.
    #          The saving interval can be changed from the `TableConfig.curriculum_save_interval_seconds`.
    run_worker(table_ip="127.0.0.1", table_port=8000)

    # 5. Stop table node
    #    Execution node type in normal use: Table
    #    NOTE: Used to explicitly terminate a table node.
    #          In this example, however, the node is running on a daemon thread,
    #          so there is no need to terminate it explicitly.
    # table_thread.stop()
    # table_thread.join()



from __future__ import annotations

from typing import TYPE_CHECKING

from lite_dist2.curriculum_models.trial import TrialStatus
from lite_dist2.study_strategies import BaseStudyStrategy, StudyStrategyModel

if TYPE_CHECKING:
    from lite_dist2.curriculum_models.trial import Mapping
    from lite_dist2.curriculum_models.trial_table import TrialTable
    from lite_dist2.value_models.aligned_space import ParameterAlignedSpace


class AllCalculationStudyStrategy(BaseStudyStrategy):
    def is_done(self, trial_table: TrialTable, parameter_space: ParameterAlignedSpace) -> bool:
        return trial_table.count_grid() == parameter_space.total

    def extract_mappings(self, trial_table: TrialTable) -> list[Mapping]:
        mappings = []
        for trial in trial_table.trials:
            if trial.trial_status != TrialStatus.done:
                continue
            mappings.extend(trial.result or [])
        return mappings

    def to_model(self) -> StudyStrategyModel:
        return StudyStrategyModel(
            type="all_calculation",
            study_strategy_param=self.study_strategy_param,
        )



import time
from pathlib import Path
from typing import Any

from lite_dist2.common import float2hex, int2hex
from lite_dist2.config import TableConfig, TableConfigProvider, WorkerConfig
from lite_dist2.curriculum_models.study_portables import StudyRegistry
from lite_dist2.study_strategies import StudyStrategyModel
from lite_dist2.suggest_strategies import SuggestStrategyModel
from lite_dist2.suggest_strategies.base_suggest_strategy import SuggestStrategyParam
from lite_dist2.table_node_api.start_table_api import start_in_thread
from lite_dist2.table_node_api.table_param import StudyRegisterParam
from lite_dist2.type_definitions import RawParamType, RawResultType
from lite_dist2.value_models.aligned_space_registry import LineSegmentRegistry, ParameterAlignedSpaceRegistry
from lite_dist2.worker_node.table_node_client import TableNodeClient
from lite_dist2.worker_node.trial_runner import AutoMPTrialRunner
from lite_dist2.worker_node.worker import Worker


def set_table_config() -> None:
    config = TableConfig(
        port=8000,
        curriculum_path=Path(__file__).parent / "example_curriculum.json",
        curriculum_save_interval_seconds=10,
    )
    TableConfigProvider.set(config)


def register_study(table_ip: str, table_port: int) -> None:
    _resolution = 10
    _half_size = 2.0

    study_register_param = StudyRegisterParam(
        study=StudyRegistry(
            name="mandelbrot",
            required_capacity=set(),
            study_strategy=StudyStrategyModel(type="all_calculation", study_strategy_param=None),
            suggest_strategy=SuggestStrategyModel(
                type="sequential",
                suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
            ),
            result_type="scalar",
            result_value_type="int",
            parameter_space=ParameterAlignedSpaceRegistry(
                type="aligned",
                axes=[
                    LineSegmentRegistry(
                        name="x",
                        type="float",
                        size=int2hex(_resolution),
                        step=float2hex(2 * _half_size / _resolution),
                        start=float2hex(-1 * _half_size),
                    ),
                    LineSegmentRegistry(
                        name="y",
                        type="float",
                        size=int2hex(_resolution),
                        step=float2hex(2 * _half_size / _resolution),
                        start=float2hex(-1 * _half_size),
                    ),
                ],
            ),
        ),
    )
    client = TableNodeClient(table_ip, table_port)
    client.register_study(study_register_param)


class Mandelbrot(AutoMPTrialRunner):
    _ABS_THRESHOLD = 2.0
    _MAX_ITER = 255

    def func(self, parameters: RawParamType, *_: tuple[Any, ...], **__: dict[str, Any]) -> RawResultType:
        x = float(parameters[0])
        y = float(parameters[1])
        c = complex(x, y)
        z = complex(0, 0)
        iter_count = 0
        while abs(z) <= self._ABS_THRESHOLD and iter_count < self._MAX_ITER:
            z = z**2 + c
            iter_count += 1
        return iter_count


def run_worker(table_ip: str, table_port: int) -> None:
    worker_config = WorkerConfig(
        name="w_01",
        process_num=2,
        max_size=10,
        wait_seconds_on_no_trial=5,
        table_node_request_timeout_seconds=60,
    )
    worker = Worker(
        trial_runner=Mandelbrot(),
        ip=table_ip,
        port=table_port,
        config=worker_config,
    )
    worker.start(stop_at_no_trial=True)


if __name__ == "__main__":
    # 1. Set table config for example
    #    Execution node type in normal use: Table
    #    NOTE: For normal use, change the `table_config.json` file in the root directory.
    set_table_config()

    # 2. Start table node in other thread
    #    Execution node type in normal use: Table
    #    NOTE: In most cases, use `start()` (or `uv run start`) instead of `start_in_thread()`
    #          since table nodes and worker nodes are separated.
    table_thread = start_in_thread()
    time.sleep(1)  # wait for activate

    # 3. Register study to table node
    #    Execution node type in normal use: Management
    #    NOTE: If the management node is also a worker node or a table node,
    #          `TableNodeClient.register_study` is available.
    #          Otherwise, studies can be registered using the curl command.
    register_study(table_ip="127.0.0.1", table_port=8000)

    # 4. run worker node
    #    Execution node type in normal use: Worker
    #    NOTE: In this example, the entire process should be completed within 10 seconds.
    #          10 seconds later, example_curriculum.json should be saved in the same directory as this file.
    #          The saving interval can be changed from the `TableConfig.curriculum_save_interval_seconds`.
    run_worker(table_ip="127.0.0.1", table_port=8000)

    # 5. Stop table node
    #    Execution node type in normal use: Table
    #    NOTE: Used to explicitly terminate a table node.
    #          In this example, however, the node is running on a daemon thread,
    #          so there is no need to terminate it explicitly.
    # table_thread.stop()
    # table_thread.join()



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING

import pytest

from lite_dist2.curriculum_models.study_portables import StudyModel, StudyRegistry, StudyStorage, StudySummary
from lite_dist2.curriculum_models.study_status import StudyStatus
from lite_dist2.curriculum_models.trial import Mapping
from lite_dist2.study_strategies import StudyStrategyModel
from lite_dist2.study_strategies.base_study_strategy import StudyStrategyParam
from lite_dist2.suggest_strategies import SuggestStrategyModel
from lite_dist2.suggest_strategies.base_suggest_strategy import SuggestStrategyParam
from lite_dist2.value_models.aligned_space import ParameterAlignedSpaceModel
from lite_dist2.value_models.aligned_space_registry import LineSegmentRegistry, ParameterAlignedSpaceRegistry
from lite_dist2.value_models.line_segment import LineSegmentModel
from lite_dist2.value_models.point import ScalarValue, VectorValue
from tests.const import DT

if TYPE_CHECKING:
    from pytest_mock.plugin import MockerFixture


@pytest.mark.parametrize(
    ("registry", "expected"),
    [
        (
            StudyRegistry(
                name="test_registry",
                required_capacity={"test"},
                study_strategy=StudyStrategyModel(
                    type="find_exact",
                    study_strategy_param=StudyStrategyParam(
                        target_value=VectorValue(
                            type="vector",
                            value_type="int",
                            name="test_target",
                            values=["0x0", "0x1", "0x2"],
                        ),
                    ),
                ),
                suggest_strategy=SuggestStrategyModel(
                    type="sequential",
                    suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
                ),
                parameter_space=ParameterAlignedSpaceRegistry(
                    type="aligned",
                    axes=[
                        LineSegmentRegistry(name="x", type="int", size="0x64", step="0x2", start="0x0"),
                        LineSegmentRegistry(name="y", type="float", size="0x64", step="0x1.0p-1", start="0x0.0p+0"),
                    ],
                ),
                result_type="vector",
                result_value_type="int",
            ),
            StudyModel(
                study_id="test_study_id",
                name="test_registry",
                required_capacity={"test"},
                status=StudyStatus.wait,
                registered_timestamp=DT,
                study_strategy=StudyStrategyModel(
                    type="find_exact",
                    study_strategy_param=StudyStrategyParam(
                        target_value=VectorValue(
                            type="vector",
                            value_type="int",
                            name="test_target",
                            values=["0x0", "0x1", "0x2"],
                        ),
                    ),
                ),
                suggest_strategy=SuggestStrategyModel(
                    type="sequential",
                    suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
                ),
                parameter_space=ParameterAlignedSpaceModel(
                    type="aligned",
                    axes=[
                        LineSegmentModel(
                            name="x",
                            type="int",
                            size="0x64",
                            step="0x2",
                            start="0x0",
                            ambient_size="0x64",
                            ambient_index="0x0",
                        ),
                        LineSegmentModel(
                            name="y",
                            type="float",
                            size="0x64",
                            step="0x1.0p-1",
                            start="0x0.0p+0",
                            ambient_size="0x64",
                            ambient_index="0x0",
                        ),
                    ],
                    check_lower_filling=True,
                ),
                result_type="vector",
                result_value_type="int",
            ),
        ),
    ],
)
def test_study_registry_to_study_model(registry: StudyRegistry, expected: StudyModel, mocker: MockerFixture) -> None:
    mocker.patch(
        "lite_dist2.curriculum_models.study_portables.StudyRegistry._publish_study_id",
        return_value="test_study_id",
    )
    mocker.patch("lite_dist2.curriculum_models.study_portables.publish_timestamp", return_value=DT)
    actual = registry.to_study_model()
    assert actual == expected


@pytest.mark.parametrize(
    ("study_registry", "expected"),
    [
        pytest.param(
            StudyRegistry(
                name="test_registry",
                required_capacity={"test"},
                study_strategy=StudyStrategyModel(
                    type="find_exact",
                    study_strategy_param=StudyStrategyParam(
                        target_value=VectorValue(
                            type="vector",
                            value_type="int",
                            name="test_target",
                            values=["0x0", "0x1", "0x2"],
                        ),
                    ),
                ),
                suggest_strategy=SuggestStrategyModel(
                    type="sequential",
                    suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
                ),
                parameter_space=ParameterAlignedSpaceRegistry(
                    type="aligned",
                    axes=[
                        LineSegmentRegistry(name="x", type="int", size="0x64", step="0x2", start="0x0"),
                        LineSegmentRegistry(name="y", type="float", size="0x64", step="0x1.0p-1", start="0x0.0p+0"),
                    ],
                ),
                result_type="vector",
                result_value_type="int",
            ),
            True,
            id="find_exact, finite: True",
        ),
        pytest.param(
            StudyRegistry(
                name="test_registry",
                required_capacity={"test"},
                study_strategy=StudyStrategyModel(
                    type="find_exact",
                    study_strategy_param=StudyStrategyParam(
                        target_value=VectorValue(
                            type="vector",
                            value_type="int",
                            name="test_target",
                            values=["0x0", "0x1", "0x2"],
                        ),
                    ),
                ),
                suggest_strategy=SuggestStrategyModel(
                    type="sequential",
                    suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
                ),
                parameter_space=ParameterAlignedSpaceRegistry(
                    type="aligned",
                    axes=[
                        LineSegmentRegistry(name="x", type="int", size=None, step="0x2", start="0x0"),
                        LineSegmentRegistry(name="y", type="float", size="0x64", step="0x1.0p-1", start="0x0.0p+0"),
                    ],
                ),
                result_type="vector",
                result_value_type="int",
            ),
            True,
            id="find_exact, infinite: True",
        ),
        pytest.param(
            StudyRegistry(
                name="test_registry",
                required_capacity={"test"},
                study_strategy=StudyStrategyModel(
                    type="all_calculation",
                    study_strategy_param=None,
                ),
                suggest_strategy=SuggestStrategyModel(
                    type="sequential",
                    suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
                ),
                parameter_space=ParameterAlignedSpaceRegistry(
                    type="aligned",
                    axes=[
                        LineSegmentRegistry(name="x", type="int", size="0x64", step="0x2", start="0x0"),
                        LineSegmentRegistry(name="y", type="float", size="0x64", step="0x1.0p-1", start="0x0.0p+0"),
                    ],
                ),
                result_type="vector",
                result_value_type="int",
            ),
            True,
            id="all_calculation, finite: True",
        ),
        pytest.param(
            StudyRegistry(
                name="test_registry",
                required_capacity={"test"},
                study_strategy=StudyStrategyModel(
                    type="all_calculation",
                    study_strategy_param=None,
                ),
                suggest_strategy=SuggestStrategyModel(
                    type="sequential",
                    suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
                ),
                parameter_space=ParameterAlignedSpaceRegistry(
                    type="aligned",
                    axes=[
                        LineSegmentRegistry(name="x", type="int", size=None, step="0x2", start="0x0"),
                        LineSegmentRegistry(name="y", type="float", size="0x64", step="0x1.0p-1", start="0x0.0p+0"),
                    ],
                ),
                result_type="vector",
                result_value_type="int",
            ),
            False,
            id="all_calculation, infinite: False",
        ),
    ],
)
def test_study_registry_is_valid(study_registry: StudyRegistry, expected: bool) -> None:
    actual = study_registry.is_valid()
    assert actual == expected


@pytest.mark.parametrize(
    ("storage", "expected"),
    [
        pytest.param(
            StudyStorage(
                study_id="test_1",
                name="test_name",
                required_capacity={"matrix"},
                registered_timestamp=DT,
                parameter_space=ParameterAlignedSpaceModel(
                    type="aligned",
                    axes=[
                        LineSegmentModel(
                            name="x",
                            type="int",
                            size="0x64",
                            step="0x2",
                            start="0x0",
                            ambient_size="0x64",
                            ambient_index="0x0",
                        ),
                        LineSegmentModel(
                            name="y",
                            type="float",
                            size="0x64",
                            step="0x1.0p-1",
                            start="0x0.0p+0",
                            ambient_size="0x64",
                            ambient_index="0x0",
                        ),
                    ],
                    check_lower_filling=True,
                ),
                done_timestamp=DT + timedelta(days=1),
                results=[
                    Mapping(
                        params=(
                            ScalarValue(type="scalar", name="x", value_type="int", value="0x0"),
                            ScalarValue(type="scalar", name="y", value_type="float", value="0x0.0p+0"),
                        ),
                        result=ScalarValue(type="scalar", value_type="bool", value=True),
                    ),
                ],
                result_type="scalar",
                result_value_type="bool",
                study_strategy=StudyStrategyModel(
                    type="find_exact",
                    study_strategy_param=StudyStrategyParam(
                        target_value=ScalarValue(
                            type="scalar",
                            value_type="bool",
                            value=True,
                        ),
                    ),
                ),
                suggest_strategy=SuggestStrategyModel(
                    type="sequential",
                    suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
                ),
                done_grids=10000,
            ),
            StudySummary(
                study_id="test_1",
                name="test_name",
                status=StudyStatus.done,
                required_capacity={"matrix"},
                registered_timestamp=DT,
                parameter_space=ParameterAlignedSpaceModel(
                    type="aligned",
                    axes=[
                        LineSegmentModel(
                            name="x",
                            type="int",
                            size="0x64",
                            step="0x2",
                            start="0x0",
                            ambient_size="0x64",
                            ambient_index="0x0",
                        ),
                        LineSegmentModel(
                            name="y",
                            type="float",
                            size="0x64",
                            step="0x1.0p-1",
                            start="0x0.0p+0",
                            ambient_size="0x64",
                            ambient_index="0x0",
                        ),
                    ],
                    check_lower_filling=True,
                ),
                result_type="scalar",
                result_value_type="bool",
                study_strategy=StudyStrategyModel(
                    type="find_exact",
                    study_strategy_param=StudyStrategyParam(
                        target_value=ScalarValue(
                            type="scalar",
                            value_type="bool",
                            value=True,
                        ),
                    ),
                ),
                suggest_strategy=SuggestStrategyModel(
                    type="sequential",
                    suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
                ),
                total_grids=10000,
                done_grids=10000,
            ),
        ),
    ],
)
def test_study_storage_to_summary(storage: StudyStorage, expected: StudySummary) -> None:
    actual = storage.to_summary()
    assert actual == expected



from __future__ import annotations

import abc
import itertools
from typing import TYPE_CHECKING

from lite_dist2.expections import LD2InvalidSpaceError
from lite_dist2.interfaces import Mergeable

if TYPE_CHECKING:
    from collections.abc import Generator, Sequence

    from lite_dist2.type_definitions import PrimitiveValueType
    from lite_dist2.value_models.aligned_space import ParameterAlignedSpace, ParameterAlignedSpaceModel
    from lite_dist2.value_models.jagged_space import ParameterJaggedSpaceModel
    from lite_dist2.value_models.point import ParamType


class ParameterSpace(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_dim(self) -> int:
        pass

    @abc.abstractmethod
    def get_total(self) -> int:
        pass

    @abc.abstractmethod
    def grid(self) -> Generator[tuple[PrimitiveValueType, ...], None, None]:
        pass

    @abc.abstractmethod
    def indexed_grid(self) -> Generator[tuple[tuple[int, PrimitiveValueType], ...], None, None]:
        pass

    @abc.abstractmethod
    def value_tuple_to_param_type(self, values: tuple[PrimitiveValueType, ...]) -> ParamType:
        pass

    @abc.abstractmethod
    def derived_by_same_ambient_space_with(self, other: ParameterSpace) -> bool:
        pass

    @abc.abstractmethod
    def to_aligned_list(self) -> list[ParameterAlignedSpace]:
        pass

    @abc.abstractmethod
    def lower_element_num_by_dim(self) -> tuple[int, ...]:
        pass

    @abc.abstractmethod
    def get_flatten_ambient_start_and_size_list(self) -> list[FlattenSegment]:
        pass

    @abc.abstractmethod
    def to_model(self) -> ParameterAlignedSpaceModel | ParameterJaggedSpaceModel:
        pass

    @staticmethod
    def get_lower_element_num_by_dim(ambient_sizes: Sequence[int]) -> tuple[int, ...]:
        # ambient_sizes = (a, b, c, d) -> lower_element_num_by_dim = (bcd, cd, d, 1)
        return tuple(
            list(
                itertools.accumulate(
                    ambient_sizes[1:][::-1],
                    lambda x, y: x * y,
                    initial=1,
                ),
            )[::-1],
        )


class FlattenSegment(Mergeable):
    def __init__(self, start: int, size: int | None) -> None:
        self.start = start
        self.size = size

    def __repr__(self) -> str:
        return f"{FlattenSegment.__name__}(start={self.start}, size={self.size})"

    def __eq__(self, other: FlattenSegment) -> bool:
        if isinstance(other, FlattenSegment):
            return (self.start == other.start) and (self.size == other.size)
        return False

    def __hash__(self) -> int:
        return hash((self.start, self.size))

    def get_start_index(self, *_: object) -> int:
        return self.start

    def can_merge(self, other: FlattenSegment, *_: object) -> bool:
        if self.start < other.start:
            smaller = self
            larger = other
        else:
            smaller = other
            larger = self

        if smaller.size is None:
            return False

        return smaller.start + smaller.size >= larger.start

    def merge(self, other: FlattenSegment, *_: object) -> FlattenSegment:
        if self.start < other.start:
            smaller = self
            larger = other
        else:
            smaller = other
            larger = self
        return FlattenSegment(smaller.start, smaller.size + larger.size)

    def next_start_index(self) -> int:
        if self.size is None:
            msg = "Cannot get next start index because size of this segment is None"
            raise LD2InvalidSpaceError(msg)
        return self.start + self.size



import pytest

from lite_dist2.curriculum_models.trial import Mapping, Trial, TrialStatus
from lite_dist2.curriculum_models.trial_table import TrialTable
from lite_dist2.expections import LD2ParameterError
from lite_dist2.suggest_strategies.base_suggest_strategy import SuggestStrategyParam
from lite_dist2.suggest_strategies.sequential_suggest_strategy import SequentialSuggestStrategy
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace
from lite_dist2.value_models.base_space import ParameterSpace
from lite_dist2.value_models.jagged_space import ParameterJaggedSpace
from lite_dist2.value_models.line_segment import DummyLineSegment, ParameterRangeInt
from lite_dist2.value_models.point import ScalarValue
from tests.const import DT


@pytest.mark.parametrize(
    ("strategy", "start", "max_num", "expected"),
    [
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=6, start=0, ambient_size=6, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            0,
            3,
            ParameterJaggedSpace(
                parameters=[(0,), (1,), (2,)],
                ambient_indices=[(0,), (1,), (2,)],
                axes_info=[DummyLineSegment(name="x", type="int", ambient_size=6, step=1)],
            ),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=6, start=0, ambient_size=6, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            3,
            3,
            ParameterJaggedSpace(
                parameters=[(3,), (4,), (5,)],
                ambient_indices=[(3,), (4,), (5,)],
                axes_info=[DummyLineSegment(name="x", type="int", ambient_size=6, step=1)],
            ),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=6, start=0, ambient_size=6, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            4,
            3,
            ParameterJaggedSpace(
                parameters=[(4,), (5,)],
                ambient_indices=[(4,), (5,)],
                axes_info=[DummyLineSegment(name="x", type="int", ambient_size=6, step=1)],
            ),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=3, start=0, ambient_size=6, ambient_index=0),
                        ParameterRangeInt(name="y", type="int", size=6, start=0, ambient_size=6, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            0,
            3,
            ParameterJaggedSpace(
                parameters=[(0, 0), (0, 1), (0, 2)],
                ambient_indices=[(0, 0), (0, 1), (0, 2)],
                axes_info=[
                    DummyLineSegment(name="x", type="int", ambient_size=6, step=1),
                    DummyLineSegment(name="y", type="int", ambient_size=6, step=1),
                ],
            ),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=3, start=0, ambient_size=6, ambient_index=0),
                        ParameterRangeInt(name="y", type="int", size=6, start=0, ambient_size=6, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            5,
            3,
            ParameterJaggedSpace(
                parameters=[(0, 5), (1, 0), (1, 1)],
                ambient_indices=[(0, 5), (1, 0), (1, 1)],
                axes_info=[
                    DummyLineSegment(name="x", type="int", ambient_size=6, step=1),
                    DummyLineSegment(name="y", type="int", ambient_size=6, step=1),
                ],
            ),
        ),
    ],
)
def test_sequential_suggest_strategy_jagged_suggest(
    strategy: SequentialSuggestStrategy,
    start: int,
    max_num: int,
    expected: ParameterJaggedSpace,
) -> None:
    actual = strategy._jagged_suggest(start, max_num)
    assert actual == expected


@pytest.mark.parametrize(
    ("strategy", "flatten_index", "expected"),
    [
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),  # 使わない
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=6, start=0, ambient_size=6, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            0,
            (1, 2, 3, 4, 5, 6),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=6, start=0, ambient_size=6, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            2,
            (3, 4, 5, 6),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=7, start=0, ambient_size=7, ambient_index=0),
                        ParameterRangeInt(type="int", size=3, start=0, ambient_size=3, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            0,
            (1, 2, 3, 6, 9, 12, 15, 18, 21),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=7, start=0, ambient_size=7, ambient_index=0),
                        ParameterRangeInt(type="int", size=3, start=0, ambient_size=3, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            1,
            (2, 3),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=7, start=0, ambient_size=7, ambient_index=0),
                        ParameterRangeInt(type="int", size=3, start=0, ambient_size=3, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            4,
            (5, 6),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            876,
            (877, 878, 879, 880),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            800,
            (801, 802, 803, 804, 805, 806, 807, 808, 809, 810, 820, 830, 840, 850, 860, 870, 880, 890, 900, 1000),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            790,
            (791, 792, 793, 794, 795, 796, 797, 798, 799, 800),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            799,
            (800,),
        ),
    ],
)
def test_sequential_suggest_strategy_generate_available_next_finite(
    strategy: SequentialSuggestStrategy,
    flatten_index: int,
    expected: tuple[int, ...],
) -> None:
    actual = strategy._generate_available_next_finite(flatten_index)
    assert actual == expected


@pytest.mark.parametrize(
    ("strategy", "flatten_index", "expected"),
    [
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=None, start=0, ambient_size=None, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            0,
            ((1,), True),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=None, start=0, ambient_size=None, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            4,
            ((5,), True),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=None, start=0, ambient_size=None, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            4,
            ((5, 6, 7, 8, 9, 10), False),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=None, start=0, ambient_size=None, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            9,
            ((10,), False),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=None, start=0, ambient_size=None, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            10,
            ((11, 12, 13, 14, 15, 16, 17, 18, 19, 20), True),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=None, start=0, ambient_size=None, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            10,
            ((11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 30, 40, 50, 60, 70, 80, 90, 100), False),
        ),
        (
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=None, start=0, ambient_size=None, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                        ParameterRangeInt(type="int", size=10, start=0, ambient_size=10, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            100,
            ((101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 120, 130, 140, 150, 160, 170, 180, 190, 200), True),
        ),
    ],
)
def test_sequential_suggest_strategy_generate_available_next_infinite(
    strategy: SequentialSuggestStrategy,
    flatten_index: int,
    expected: tuple[tuple[int, ...], bool],
) -> None:
    actual = strategy._generate_available_next_infinite(flatten_index)
    assert actual == expected


@pytest.mark.parametrize(
    ("init", "ratio", "iter_num", "expected"),
    [
        ((1, 2, 3), 3, 5, (1, 2, 3, 6, 9)),
        ((1, 2, 3, 4), 10, 3, (1, 2, 3)),
    ],
)
def test_sequential_suggest_strategy_infinite_available_generator(
    init: tuple[int, ...],
    ratio: int,
    iter_num: int,
    expected: tuple[int, ...],
) -> None:
    generator = SequentialSuggestStrategy._infinite_available_generator(init, ratio)
    actual = []
    for i, v in enumerate(generator):
        if i >= iter_num:
            break
        actual.append(v)
    actual = tuple(actual)
    assert actual == expected


@pytest.mark.parametrize(
    ("a", "b", "expected"),
    [
        (1, 2, 1),
        (None, 2, 2),
        (3, None, 3),
    ],
)
def test_sequential_suggest_strategy_nullable_min(a: int | None, b: int | None, expected: int) -> None:
    actual = SequentialSuggestStrategy._nullable_min(a, b)
    assert actual == expected


def test_sequential_suggest_strategy_nullable_min_raise_both_none() -> None:
    with pytest.raises(LD2ParameterError, match=r"both\sis\sNone"):
        _ = SequentialSuggestStrategy._nullable_min(None, None)


@pytest.mark.parametrize(
    ("strategy", "trial_table", "max_num", "expected"),
    [
        pytest.param(
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=False),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=6, start=0, ambient_size=6, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            TrialTable(
                trials=[],
                aggregated_parameter_space={-1: [], 0: []},
            ),
            3,
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=3, start=0, ambient_size=6, ambient_index=0),
                ],
                check_lower_filling=True,
            ),
            id="1D init",
        ),
        pytest.param(
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=False),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=6, start=0, ambient_size=6, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            TrialTable(
                trials=[
                    Trial(
                        study_id="01",
                        trial_id="01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=3,
                                    start=0,
                                    ambient_size=6,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),),
                                result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                            ),
                            Mapping(
                                params=(ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),),
                                result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                            ),
                            Mapping(
                                params=(ScalarValue(type="scalar", value_type="int", value="0x2", name="x"),),
                                result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                            ),
                        ],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=3,
                                    start=0,
                                    ambient_size=6,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            3,
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=3, start=3, ambient_size=6, ambient_index=3),
                ],
                check_lower_filling=True,
            ),
            id="1D continuing",
        ),
        pytest.param(
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=False),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(
                            name="x",
                            type="int",
                            size=None,
                            start=0,
                            ambient_size=None,
                            ambient_index=0,
                        ),
                    ],
                    check_lower_filling=True,
                ),
            ),
            TrialTable(
                trials=[],
                aggregated_parameter_space={-1: [], 0: []},
            ),
            3,
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=3, start=0, ambient_size=None, ambient_index=0),
                ],
                check_lower_filling=True,
            ),
            id="1D init infinite",
        ),
        pytest.param(
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=False),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(
                            name="x",
                            type="int",
                            size=None,
                            start=0,
                            ambient_size=None,
                            ambient_index=0,
                        ),
                    ],
                    check_lower_filling=True,
                ),
            ),
            TrialTable(
                trials=[
                    Trial(
                        study_id="01",
                        trial_id="01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=3,
                                    start=0,
                                    ambient_size=None,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),),
                                result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                            ),
                            Mapping(
                                params=(ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),),
                                result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                            ),
                            Mapping(
                                params=(ScalarValue(type="scalar", value_type="int", value="0x2", name="x"),),
                                result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                            ),
                        ],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=3,
                                    start=0,
                                    ambient_size=None,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            3,
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=3, start=3, ambient_size=None, ambient_index=3),
                ],
                check_lower_filling=True,
            ),
            id="1D continuing infinite",
        ),
        pytest.param(
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=False),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=6, start=0, ambient_size=6, ambient_index=0),
                        ParameterRangeInt(name="y", type="int", size=4, start=0, ambient_size=4, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            TrialTable(
                trials=[
                    Trial(
                        study_id="01",
                        trial_id="01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterJaggedSpace(
                            parameters=[(0, 0), (0, 1), (0, 2)],
                            ambient_indices=[(0, 0), (0, 1), (0, 2)],
                            axes_info=[
                                DummyLineSegment(name="x", type="int", ambient_size=6, step=1),
                                DummyLineSegment(name="y", type="int", ambient_size=4, step=1),
                            ],
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x2", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                            ),
                        ],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=0,
                                    ambient_size=6,
                                    ambient_index=0,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=3,
                                    start=0,
                                    ambient_size=4,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            4,
            ParameterJaggedSpace(
                parameters=[(0, 3), (1, 0), (1, 1), (1, 2)],
                ambient_indices=[(0, 3), (1, 0), (1, 1), (1, 2)],
                axes_info=[
                    DummyLineSegment(name="x", type="int", ambient_size=6, step=1),
                    DummyLineSegment(name="y", type="int", ambient_size=4, step=1),
                ],
            ),
            id="2D continuing",
        ),
        pytest.param(
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=False),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=6, start=0, ambient_size=6, ambient_index=0),
                        ParameterRangeInt(name="y", type="int", size=4, start=0, ambient_size=4, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            TrialTable(
                trials=[
                    Trial(
                        study_id="01",
                        trial_id="01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterJaggedSpace(
                            parameters=[(0, 0), (0, 1), (0, 2)],
                            ambient_indices=[(0, 0), (0, 1), (0, 2)],
                            axes_info=[
                                DummyLineSegment(name="x", type="int", ambient_size=6, step=1),
                                DummyLineSegment(name="y", type="int", ambient_size=4, step=1),
                            ],
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x2", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="01",
                        trial_id="02",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterJaggedSpace(
                            parameters=[(1, 1), (1, 2), (1, 3)],
                            ambient_indices=[(1, 1), (1, 2), (1, 3)],
                            axes_info=[
                                DummyLineSegment(name="x", type="int", ambient_size=6, step=1),
                                DummyLineSegment(name="y", type="int", ambient_size=4, step=1),
                            ],
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x69"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x2", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x6a"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x3", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x6b"),
                            ),
                        ],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=0,
                                    ambient_size=6,
                                    ambient_index=0,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=3,
                                    start=0,
                                    ambient_size=4,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=1,
                                    ambient_size=6,
                                    ambient_index=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=3,
                                    start=1,
                                    ambient_size=4,
                                    ambient_index=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            4,
            ParameterJaggedSpace(
                parameters=[(0, 3), (1, 0)],
                ambient_indices=[(0, 3), (1, 0)],
                axes_info=[
                    DummyLineSegment(name="x", type="int", ambient_size=6, step=1),
                    DummyLineSegment(name="y", type="int", ambient_size=4, step=1),
                ],
            ),
            id="2D striping",
        ),
        pytest.param(
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=6, start=0, ambient_size=6, ambient_index=0),
                        ParameterRangeInt(name="y", type="int", size=4, start=0, ambient_size=4, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            TrialTable(
                trials=[
                    Trial(
                        study_id="01",
                        trial_id="01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=0,
                                    ambient_size=6,
                                    ambient_index=0,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=3,
                                    start=0,
                                    ambient_size=4,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x2", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="01",
                        trial_id="02",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=1,
                                    ambient_size=6,
                                    ambient_index=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=3,
                                    start=1,
                                    ambient_size=4,
                                    ambient_index=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x69"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x2", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x6a"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x3", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x6b"),
                            ),
                        ],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=0,
                                    ambient_size=6,
                                    ambient_index=0,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=3,
                                    start=0,
                                    ambient_size=4,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=1,
                                    ambient_size=6,
                                    ambient_index=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=3,
                                    start=1,
                                    ambient_size=4,
                                    ambient_index=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            4,
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=1, start=0, ambient_size=6, ambient_index=0),
                    ParameterRangeInt(name="y", type="int", size=1, start=3, ambient_size=4, ambient_index=3),
                ],
                check_lower_filling=True,
            ),
            id="2D striping force aligned",
        ),
        pytest.param(
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=False),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(
                            name="x",
                            type="int",
                            size=None,
                            start=0,
                            ambient_size=None,
                            ambient_index=0,
                        ),
                        ParameterRangeInt(
                            name="y",
                            type="int",
                            size=4,
                            start=0,
                            ambient_size=4,
                            ambient_index=0,
                        ),
                    ],
                    check_lower_filling=True,
                ),
            ),
            TrialTable(
                trials=[
                    Trial(
                        study_id="01",
                        trial_id="01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterJaggedSpace(
                            parameters=[(0, 0), (0, 1), (0, 2)],
                            ambient_indices=[(0, 0), (0, 1), (0, 2)],
                            axes_info=[
                                DummyLineSegment(name="x", type="int", ambient_size=6, step=1),
                                DummyLineSegment(name="y", type="int", ambient_size=4, step=1),
                            ],
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x2", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                            ),
                        ],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=0,
                                    ambient_size=None,
                                    ambient_index=0,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=3,
                                    start=0,
                                    ambient_size=4,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            4,
            ParameterJaggedSpace(
                parameters=[(0, 3), (1, 0), (1, 1), (1, 2)],
                ambient_indices=[(0, 3), (1, 0), (1, 1), (1, 2)],
                axes_info=[
                    DummyLineSegment(name="x", type="int", ambient_size=None, step=1),
                    DummyLineSegment(name="y", type="int", ambient_size=4, step=1),
                ],
            ),
            id="2D continuing infinite",
        ),
        pytest.param(
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=2, start=0, ambient_size=2, ambient_index=0),
                        ParameterRangeInt(name="y", type="int", size=2, start=0, ambient_size=2, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            TrialTable(
                trials=[
                    Trial(
                        study_id="01",
                        trial_id="01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=2,
                                    start=0,
                                    ambient_size=2,
                                    ambient_index=0,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=2,
                                    start=0,
                                    ambient_size=2,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x67"),
                            ),
                        ],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=0,
                                    ambient_size=2,
                                    ambient_index=0,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=2,
                                    start=0,
                                    ambient_size=2,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=1,
                                    ambient_size=2,
                                    ambient_index=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=2,
                                    start=0,
                                    ambient_size=2,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                    1: [],
                },
            ),
            4,
            None,
            id="2D full",
        ),
        pytest.param(
            SequentialSuggestStrategy(
                suggest_parameter=SuggestStrategyParam(strict_aligned=True),
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=51, start=0, ambient_size=51, ambient_index=0),
                        ParameterRangeInt(name="y", type="int", size=51, start=0, ambient_size=51, ambient_index=0),
                    ],
                    check_lower_filling=True,
                ),
            ),
            TrialTable(
                trials=[
                    Trial(
                        study_id="01",
                        trial_id="00",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=0,
                                    ambient_size=51,
                                    ambient_index=0,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=51,
                                    start=0,
                                    ambient_size=51,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[],
                    ),
                    Trial(
                        study_id="01",
                        trial_id="01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=1,
                                    ambient_size=51,
                                    ambient_index=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=51,
                                    start=0,
                                    ambient_size=51,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w02",
                        worker_node_id="w02",
                        results=[],
                    ),
                    Trial(
                        study_id="01",
                        trial_id="02",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=2,
                                    ambient_size=51,
                                    ambient_index=2,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=51,
                                    start=0,
                                    ambient_size=51,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[],
                    ),
                    Trial(
                        study_id="01",
                        trial_id="03",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=3,
                                    ambient_size=51,
                                    ambient_index=3,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=51,
                                    start=0,
                                    ambient_size=51,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[],
                    ),
                    Trial(
                        study_id="01",
                        trial_id="04",
                        timestamp=DT,
                        trial_status=TrialStatus.running,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=4,
                                    ambient_size=51,
                                    ambient_index=4,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=51,
                                    start=0,
                                    ambient_size=51,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w02",
                        worker_node_id="w02",
                        results=None,
                    ),
                    Trial(
                        study_id="01",
                        trial_id="05",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=5,
                                    ambient_size=51,
                                    ambient_index=5,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=51,
                                    start=0,
                                    ambient_size=51,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[],
                    ),
                    Trial(
                        study_id="01",
                        trial_id="06",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=6,
                                    ambient_size=51,
                                    ambient_index=6,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=51,
                                    start=0,
                                    ambient_size=51,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=6,
                                    ambient_size=51,
                                    ambient_index=6,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=51,
                                    start=0,
                                    ambient_size=51,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=4,
                                    start=0,
                                    ambient_size=51,
                                    ambient_index=0,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=51,
                                    start=0,
                                    ambient_size=51,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    start=5,
                                    ambient_size=51,
                                    ambient_index=5,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=51,
                                    start=0,
                                    ambient_size=51,
                                    ambient_index=0,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                    1: [],
                },
            ),
            51,
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=1, start=7, ambient_size=51, ambient_index=7),
                    ParameterRangeInt(name="y", type="int", size=51, start=0, ambient_size=51, ambient_index=0),
                ],
                check_lower_filling=True,
            ),
            id="2D 2 node",
        ),
    ],
)
def test_sequential_suggest_strategy_suggest(
    strategy: SequentialSuggestStrategy,
    trial_table: TrialTable,
    max_num: int,
    expected: ParameterSpace | None,
) -> None:
    actual = strategy.suggest(trial_table, max_num)
    assert actual == expected



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from lite_dist2.common import hex2float
from lite_dist2.expections import LD2ParameterError
from lite_dist2.value_models.line_segment import (
    LineSegment,
    LineSegmentModel,
    ParameterRangeBool,
    ParameterRangeFloat,
    ParameterRangeInt,
)

if TYPE_CHECKING:
    from lite_dist2.type_definitions import PrimitiveValueType


@pytest.mark.parametrize(
    ("seg", "expected"),
    [
        (ParameterRangeBool(type="bool", size=1, ambient_index=0, ambient_size=2, start=False), 0),
        (ParameterRangeBool(type="bool", size=1, ambient_index=1, ambient_size=1, start=True), 1),
        (ParameterRangeInt(type="int", size=10, ambient_index=0, ambient_size=255, start=0), 9),
        (ParameterRangeInt(type="int", size=10, ambient_index=5, ambient_size=255, start=0), 14),
        (ParameterRangeFloat(type="float", size=5, ambient_index=0, ambient_size=255, start=11, step=1), 4),
        (ParameterRangeFloat(type="float", size=5, ambient_index=2, ambient_size=255, start=11, step=1), 6),
    ],
)
def test_line_segment_end_index(seg: LineSegment, expected: int) -> None:
    actual = seg.end_index()
    assert actual == expected


@pytest.mark.parametrize(
    ("one", "other", "expected"),
    [
        (  # complete match True
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=255, start=0, step=1),
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=255, start=0, step=1),
            True,
        ),
        (  # different start True
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=255, start=0, step=1),
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=16, ambient_size=255, start=10, step=1),
            True,
        ),
        (  # different size True
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=255, start=0, step=1),
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=16, ambient_size=255, start=10, step=1),
            True,
        ),
        (  # different name False
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=255, start=0, step=1),
            ParameterRangeInt(name="y", type="int", size=100, ambient_index=16, ambient_size=255, start=10, step=1),
            False,
        ),
        (  # different type False
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=255, start=0, step=1),
            ParameterRangeBool(name="x", type="bool", size=1, ambient_index=0, ambient_size=2, start=True, step=1),
            False,
        ),
        (  # different ambient_size False
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=255, start=0, step=1),
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=16, ambient_size=16, start=10, step=1),
            False,
        ),
        (  # different step False
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=255, start=0, step=1),
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=16, ambient_size=255, start=10, step=2),
            False,
        ),
    ],
)
def test_line_segment_derived_by_same_ambient_space_with(one: LineSegment, other: LineSegment, expected: bool) -> None:
    actual = one.derived_by_same_ambient_space_with(other)
    actual_reverse = other.derived_by_same_ambient_space_with(one)
    assert actual == expected
    assert actual_reverse == expected


@pytest.mark.parametrize(
    ("one", "other", "expected"),
    [  # Cases that can be excluded by `derived_by_same_ambient_space_with` are not test here
        (  # complete match True
            ParameterRangeBool(name="x", type="bool", size=1, ambient_index=0, ambient_size=2, start=False),
            ParameterRangeBool(name="x", type="bool", size=1, ambient_index=0, ambient_size=2, start=False),
            True,
        ),
        (  # adjacency True
            ParameterRangeBool(name="x", type="bool", size=1, ambient_index=0, ambient_size=2, start=False),
            ParameterRangeBool(name="x", type="bool", size=1, ambient_index=1, ambient_size=2, start=True),
            True,
        ),
        (  # complete match True
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=0, ambient_size=255, start=0),
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=0, ambient_size=255, start=0),
            True,
        ),
        (  # adjacency True
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=0, ambient_size=255, start=0),
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=10, ambient_size=255, start=10),
            True,
        ),
        (  # overlap True
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=0, ambient_size=255, start=0),
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=5, ambient_size=255, start=5),
            True,
        ),
        (  # stride False
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=0, ambient_size=255, start=0),
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=11, ambient_size=255, start=11),
            False,
        ),
        (  # Complete match True
            ParameterRangeFloat(type="float", size=10, ambient_index=0, ambient_size=255, start=0.0, step=1),
            ParameterRangeFloat(type="float", size=10, ambient_index=0, ambient_size=255, start=0.0, step=1),
            True,
        ),
        (  # overlap True
            ParameterRangeFloat(type="float", size=10, ambient_index=0, ambient_size=255, start=0.0, step=1),
            ParameterRangeFloat(type="float", size=10, ambient_index=5, ambient_size=255, start=5.0, step=1),
            True,
        ),
        (  # adjacency True
            ParameterRangeFloat(type="float", size=10, ambient_index=0, ambient_size=255, start=0.0, step=1),
            ParameterRangeFloat(type="float", size=10, ambient_index=10, ambient_size=255, start=10.0, step=1),
            True,
        ),
        (  # stride False
            ParameterRangeFloat(type="float", size=10, ambient_index=0, ambient_size=255, start=0.0, step=1),
            ParameterRangeFloat(type="float", size=10, ambient_index=11, ambient_size=255, start=11.0, step=1),
            False,
        ),
    ],
)
def test_line_segment_can_merge(one: LineSegment, other: LineSegment, expected: bool) -> None:
    actual = one.can_merge(other)
    actual_reverse = other.can_merge(one)
    assert actual == expected
    assert actual_reverse == expected


@pytest.mark.parametrize(
    ("one", "other", "expected"),
    [
        (  # complete match True
            ParameterRangeBool(name="x", type="bool", size=1, ambient_index=0, ambient_size=2, start=False),
            ParameterRangeBool(name="x", type="bool", size=1, ambient_index=0, ambient_size=2, start=False),
            ParameterRangeBool(name="x", type="bool", size=1, ambient_index=0, ambient_size=2, start=False),
        ),
        (  # adjacency True
            ParameterRangeBool(name="x", type="bool", size=1, ambient_index=0, ambient_size=2, start=False),
            ParameterRangeBool(name="x", type="bool", size=1, ambient_index=1, ambient_size=2, start=True),
            ParameterRangeBool(name="x", type="bool", size=2, ambient_index=0, ambient_size=2, start=False),
        ),
        (  # complete match True
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=0, ambient_size=255, start=0),
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=0, ambient_size=255, start=0),
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=0, ambient_size=255, start=0),
        ),
        (  # adjacency True
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=0, ambient_size=255, start=0),
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=10, ambient_size=255, start=10),
            ParameterRangeInt(name="x", type="int", size=20, ambient_index=0, ambient_size=255, start=0),
        ),
        (  # overlap True
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=0, ambient_size=255, start=0),
            ParameterRangeInt(name="x", type="int", size=10, ambient_index=5, ambient_size=255, start=5),
            ParameterRangeInt(name="x", type="int", size=15, ambient_index=0, ambient_size=255, start=0),
        ),
        (  # Complete match True
            ParameterRangeFloat(type="float", size=10, ambient_index=0, ambient_size=255, start=0.0, step=1),
            ParameterRangeFloat(type="float", size=10, ambient_index=0, ambient_size=255, start=0.0, step=1),
            ParameterRangeFloat(type="float", size=10, ambient_index=0, ambient_size=255, start=0.0, step=1),
        ),
        (  # overlap True
            ParameterRangeFloat(type="float", size=10, ambient_index=0, ambient_size=255, start=0.0, step=1),
            ParameterRangeFloat(type="float", size=10, ambient_index=5, ambient_size=255, start=5.0, step=1),
            ParameterRangeFloat(type="float", size=15, ambient_index=0, ambient_size=255, start=0.0, step=1),
        ),
        (  # adjacency True
            ParameterRangeFloat(type="float", size=10, ambient_index=0, ambient_size=255, start=0.0, step=1),
            ParameterRangeFloat(type="float", size=10, ambient_index=10, ambient_size=255, start=10.0, step=1),
            ParameterRangeFloat(type="float", size=20, ambient_index=0, ambient_size=255, start=0.0, step=1),
        ),
    ],
)
def test_line_segment_merge(one: LineSegment, other: LineSegment, expected: LineSegment) -> None:
    actual = one.merge(other)
    actual_reverse = other.merge(one)
    assert actual == expected
    assert actual_reverse == expected


@pytest.mark.parametrize(
    ("line_segment", "expected"),
    [
        (ParameterRangeBool(type="bool", size=1, ambient_index=0, ambient_size=1, start=False), True),
        (ParameterRangeBool(type="bool", size=2, ambient_index=0, ambient_size=2, start=False), True),
        (ParameterRangeBool(type="bool", size=1, ambient_index=0, ambient_size=2, start=True), False),
        (ParameterRangeInt(type="int", size=10, ambient_index=0, ambient_size=100, start=0), False),
        (ParameterRangeInt(type="int", size=10, ambient_index=0, ambient_size=10, start=0), True),
        (ParameterRangeFloat(type="float", size=10, ambient_index=0, ambient_size=100, start=0, step=2.0), False),
        (ParameterRangeFloat(type="float", size=10, ambient_index=0, ambient_size=10, start=0, step=2.0), True),
    ],
)
def test_line_segment_is_universal(line_segment: LineSegment, expected: bool) -> None:
    actual = line_segment.is_universal()
    assert actual == expected


@pytest.mark.parametrize(
    ("line_segment", "start_index", "size", "expected"),
    [
        (
            ParameterRangeBool(name="x", type="bool", size=2, ambient_index=0, ambient_size=2, start=False),
            0,
            1,
            ParameterRangeBool(name="x", type="bool", size=1, ambient_index=0, ambient_size=2, start=False),
        ),
        (
            ParameterRangeBool(name="x", type="bool", size=2, ambient_index=0, ambient_size=2, start=False),
            1,
            1,
            ParameterRangeBool(name="x", type="bool", size=1, ambient_index=1, ambient_size=2, start=True),
        ),
        (
            ParameterRangeBool(name="x", type="bool", size=2, ambient_index=0, ambient_size=2, start=False),
            0,
            2,
            ParameterRangeBool(name="x", type="bool", size=2, ambient_index=0, ambient_size=2, start=False),
        ),
        (
            ParameterRangeInt(type="int", size=10, ambient_index=20, ambient_size=100, start=20, step=1),
            5,
            4,
            ParameterRangeInt(type="int", size=4, ambient_index=25, ambient_size=100, start=25, step=1),
        ),
        (
            ParameterRangeInt(type="int", size=10, ambient_index=20, ambient_size=100, start=20, step=1),
            0,
            10,
            ParameterRangeInt(type="int", size=10, ambient_index=20, ambient_size=100, start=20, step=1),
        ),
        (
            ParameterRangeFloat(type="float", size=10, ambient_index=20, ambient_size=100, start=-100.0, step=5.0),
            5,
            4,
            ParameterRangeFloat(type="float", size=4, ambient_index=25, ambient_size=100, start=-75.0, step=5.0),
        ),
        (
            ParameterRangeFloat(type="float", size=10, ambient_index=20, ambient_size=100, start=-100.0, step=5.0),
            0,
            10,
            ParameterRangeFloat(type="float", size=10, ambient_index=20, ambient_size=100, start=-100.0, step=5.0),
        ),
    ],
)
def test_line_segment_slice(line_segment: LineSegment, start_index: int, size: int, expected: LineSegment) -> None:
    actual = line_segment.slice(start_index, size)
    assert actual == expected


@pytest.mark.parametrize(
    ("line_segment", "size"),
    [
        (ParameterRangeBool(type="bool", size=2, ambient_index=0, ambient_size=2, start=False), 3),
        (ParameterRangeInt(type="int", size=10, ambient_index=20, ambient_size=100, start=20, step=1), 11),
        (ParameterRangeFloat(type="float", size=10, ambient_index=20, ambient_size=100, start=-100.0, step=5.0), 11),
    ],
)
def test_line_segment_slice_raise(line_segment: LineSegment, size: int) -> None:
    start_index = 0
    with pytest.raises(LD2ParameterError):
        _ = line_segment.slice(start_index, size)


@pytest.mark.parametrize(
    ("line_segment", "expected"),
    [
        pytest.param(
            ParameterRangeBool(type="bool", size=2, ambient_index=0, ambient_size=2, start=False),
            [False, True],
            id="bool",
        ),
        pytest.param(
            ParameterRangeInt(type="int", size=10, ambient_index=40, ambient_size=100, start=20, step=1),
            [20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
            id="int",
        ),
        pytest.param(
            ParameterRangeFloat(type="float", size=5, ambient_index=40, ambient_size=100, start=20.0, step=0.5),
            [20.0, 20.5, 21.0, 21.5, 22.0],
            id="float",
        ),
    ],
)
def test_line_segment_grid_finite(line_segment: LineSegment, expected: list[PrimitiveValueType]) -> None:
    actual = list(line_segment.grid())
    assert actual == expected


@pytest.mark.parametrize(
    ("line_segment", "max_num", "expected"),
    [
        pytest.param(
            ParameterRangeInt(type="int", size=None, ambient_index=20, ambient_size=None, start=20, step=1),
            5,
            [20, 21, 22, 23, 24],
            id="int",
        ),
        pytest.param(
            ParameterRangeFloat(type="float", size=None, ambient_index=40, ambient_size=None, start=20.0, step=0.5),
            5,
            [20.0, 20.5, 21.0, 21.5, 22.0],
            id="float",
        ),
    ],
)
def test_line_segment_grid_infinite(
    line_segment: LineSegment,
    max_num: int,
    expected: list[PrimitiveValueType],
) -> None:
    actual = []
    for item in line_segment.grid():
        actual.append(item)
        if len(actual) >= max_num:
            break
    assert actual == expected


@pytest.mark.parametrize(
    ("line_segment", "expected"),
    [
        pytest.param(
            ParameterRangeBool(type="bool", size=2, ambient_index=0, ambient_size=2, start=False),
            [(0, False), (1, True)],
            id="bool",
        ),
        pytest.param(
            ParameterRangeInt(type="int", size=10, ambient_index=40, ambient_size=100, start=20, step=1),
            [(40, 20), (41, 21), (42, 22), (43, 23), (44, 24), (45, 25), (46, 26), (47, 27), (48, 28), (49, 29)],
            id="int",
        ),
        pytest.param(
            ParameterRangeFloat(type="float", size=5, ambient_index=40, ambient_size=100, start=20.0, step=0.5),
            [(40, 20.0), (41, 20.5), (42, 21.0), (43, 21.5), (44, 22.0)],
            id="float",
        ),
    ],
)
def test_line_segment_indexed_grid_finite(
    line_segment: LineSegment,
    expected: list[tuple[int, PrimitiveValueType]],
) -> None:
    actual = list(line_segment.indexed_grid())
    assert actual == expected


@pytest.mark.parametrize(
    ("line_segment", "max_num", "expected"),
    [
        pytest.param(
            ParameterRangeInt(type="int", size=None, ambient_index=40, ambient_size=None, start=20, step=1),
            5,
            [(40, 20), (41, 21), (42, 22), (43, 23), (44, 24)],
            id="int",
        ),
        pytest.param(
            ParameterRangeFloat(type="float", size=None, ambient_index=40, ambient_size=None, start=20.0, step=0.5),
            5,
            [(40, 20.0), (41, 20.5), (42, 21.0), (43, 21.5), (44, 22.0)],
            id="float",
        ),
    ],
)
def test_line_segment_indexed_grid_infinite(
    line_segment: LineSegment,
    max_num: int,
    expected: list[tuple[int, PrimitiveValueType]],
) -> None:
    actual = []
    for item in line_segment.indexed_grid():
        actual.append(item)
        if len(actual) >= max_num:
            break
    assert actual == expected


def test_line_segment_grid_without_rounding_error() -> None:
    ambient = ParameterRangeFloat(
        type="float",
        size=13,
        step=0.5 / 13,
        start=-0.25,
        ambient_index=0,
        ambient_size=13,
    )
    ambient_grid = list(ambient.grid())

    derived = ParameterRangeFloat(
        type="float",
        size=6,
        step=0.5 / 13,
        start=0.5 / 13 * 7 - 0.25,
        ambient_index=7,
        ambient_size=13,
    )
    derived_grid = list(derived.grid())

    amb_idx = 7
    assert len(ambient_grid[amb_idx:]) == len(derived_grid)
    for i in range(5):
        assert ambient_grid[i + amb_idx] == derived_grid[i]


@pytest.mark.parametrize(
    "model",
    [
        LineSegmentModel(
            name="x",
            type="bool",
            size="0x2",
            step="0x1",
            start=False,
            ambient_index="0x0",
            ambient_size="0x2",
        ),
        LineSegmentModel(
            type="bool",
            size="0x2",
            step="0x1",
            start=False,
            ambient_index="0x0",
            ambient_size="0x2",
        ),
    ],
)
def test_line_segment_model_bool_to_model_from_model(model: LineSegmentModel) -> None:
    segment = ParameterRangeBool.from_model(model)
    reconstructed_model = segment.to_model()
    assert model == reconstructed_model


@pytest.mark.parametrize(
    "model",
    [
        LineSegmentModel(
            name="x",
            type="int",
            size="0xa",
            step="0x1",
            start="0x0",
            ambient_index="0x0",
            ambient_size="0x64",
        ),
        LineSegmentModel(
            type="int",
            size="0xa",
            step="0x1",
            start="0x0",
            ambient_index="0x0",
            ambient_size="0x64",
        ),
    ],
)
def test_line_segment_model_int_to_model_from_model(model: LineSegmentModel) -> None:
    segment = ParameterRangeInt.from_model(model)
    reconstructed_model = segment.to_model()
    assert model == reconstructed_model


@pytest.mark.parametrize(
    "model",
    [
        LineSegmentModel(
            name="x",
            type="float",
            size="0xa",
            step="0x1.0p-1",
            start="0x0.0p+0",
            ambient_index="0x0",
            ambient_size="0x64",
        ),
        LineSegmentModel(
            type="float",
            size="0xa",
            step="0x1.0p-1",
            start="0x0.0p+0",
            ambient_index="0x0",
            ambient_size="0x64",
        ),
    ],
)
def test_line_segment_model_float_to_model_from_model(model: LineSegmentModel) -> None:
    segment = ParameterRangeFloat.from_model(model)
    reconstructed_model = segment.to_model()
    assert model.name == reconstructed_model.name
    assert model.type == reconstructed_model.type
    assert model.size == reconstructed_model.size
    assert model.ambient_index == reconstructed_model.ambient_index
    assert model.ambient_size == reconstructed_model.ambient_size
    assert hex2float(model.start) == hex2float(reconstructed_model.start)
    assert hex2float(model.step) == hex2float(reconstructed_model.step)



from __future__ import annotations

import uuid
from datetime import datetime
from typing import Literal

from pydantic import BaseModel, Field

from lite_dist2.common import publish_timestamp
from lite_dist2.curriculum_models.study_status import StudyStatus
from lite_dist2.curriculum_models.trial import Mapping
from lite_dist2.curriculum_models.trial_table import TrialTableModel
from lite_dist2.study_strategies import StudyStrategyModel
from lite_dist2.suggest_strategies import SuggestStrategyModel
from lite_dist2.value_models.aligned_space import ParameterAlignedSpaceModel
from lite_dist2.value_models.aligned_space_registry import ParameterAlignedSpaceRegistry


class _StudyCommonModel(BaseModel):
    name: str | None
    required_capacity: set[str]
    study_strategy: StudyStrategyModel
    suggest_strategy: SuggestStrategyModel
    result_type: Literal["scalar", "vector"]
    result_value_type: Literal["bool", "int", "float"]


class StudyModel(_StudyCommonModel):
    """
    For save full information
    """

    study_id: str
    status: StudyStatus
    registered_timestamp: datetime
    parameter_space: ParameterAlignedSpaceModel
    trial_table: TrialTableModel = Field(default_factory=TrialTableModel.create_empty)


class StudyRegistry(_StudyCommonModel):
    """
    For registration to curriculum
    """

    parameter_space: ParameterAlignedSpaceRegistry

    def is_valid(self) -> bool:
        is_infinite = any(axis.size is None for axis in self.parameter_space.axes)
        return not (is_infinite and self.study_strategy.type == "all_calculation")

    def to_study_model(self) -> StudyModel:
        return StudyModel(
            study_id=self._publish_study_id(),
            name=self.name,
            required_capacity=self.required_capacity,
            status=StudyStatus.wait,
            registered_timestamp=publish_timestamp(),
            study_strategy=self.study_strategy,
            suggest_strategy=self.suggest_strategy,
            parameter_space=self.parameter_space.to_parameter_aligned_space_model(),
            result_type=self.result_type,
            result_value_type=self.result_value_type,
        )

    def _publish_study_id(self) -> str:
        node = hash(self.name) if self.name is not None else hash(self.required_capacity)
        if isinstance(node, int):
            # 下 48 bit だけ取得
            node = node & 0xFFFFFFFFFFFF
        return str(uuid.uuid1(node=node, clock_seq=None))


class StudySummary(_StudyCommonModel):
    """
    For showing summary status
    """

    study_id: str
    status: StudyStatus
    registered_timestamp: datetime
    parameter_space: ParameterAlignedSpaceModel
    total_grids: int | None
    done_grids: int


class StudyStorage(_StudyCommonModel):
    """
    For conclusional result
    """

    study_id: str
    registered_timestamp: datetime
    parameter_space: ParameterAlignedSpaceModel
    done_timestamp: datetime
    results: list[Mapping]
    done_grids: int

    def to_summary(self) -> StudySummary:
        return StudySummary(
            name=self.name,
            study_id=self.study_id,
            required_capacity=self.required_capacity,
            status=StudyStatus.done,
            registered_timestamp=self.registered_timestamp,
            study_strategy=self.study_strategy,
            suggest_strategy=self.suggest_strategy,
            parameter_space=self.parameter_space,
            result_type=self.result_type,
            result_value_type=self.result_value_type,
            total_grids=self.parameter_space.get_total(),
            done_grids=self.done_grids,
        )



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

from lite_dist2.common import numerize, portablize
from lite_dist2.type_definitions import PortableValueType

if TYPE_CHECKING:
    from collections.abc import Iterable

    from lite_dist2.type_definitions import PrimitiveValueType, RawResultType


class BasePointValue(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def numerize(self) -> PrimitiveValueType | list[PrimitiveValueType]:
        pass

    @abc.abstractmethod
    def equal_to(self, other: BasePointValue) -> bool:
        pass

    @staticmethod
    @abc.abstractmethod
    def create_from_numeric(
        raw_result_value: RawResultType,
        value_type: Literal["bool", "int", "float"],
        name: str | None = None,
    ) -> BasePointValue:
        pass


class ScalarValue(BaseModel, BasePointValue):
    type: Literal["scalar"]
    value_type: Literal["bool", "int", "float"]
    value: PortableValueType
    name: str | None = None

    def numerize(self) -> PrimitiveValueType:
        return numerize(self.value_type, self.value)

    def equal_to(self, other: BasePointValue) -> bool:
        if not isinstance(other, ScalarValue):
            return False
        return self.value_type == other.value_type and self.value == other.value

    @staticmethod
    def create_from_numeric(
        raw_result_value: RawResultType,
        value_type: Literal["bool", "int", "float"],
        name: str | None = None,
    ) -> ScalarValue:
        val = portablize(value_type, raw_result_value)
        return ScalarValue(type="scalar", value_type=value_type, value=val, name=name)


class VectorValue(BaseModel, BasePointValue):
    type: Literal["vector"]
    value_type: Literal["bool", "int", "float"]
    values: list[PortableValueType]
    name: str | None = None

    def numerize(self) -> list[PrimitiveValueType]:
        return [numerize(self.value_type, v) for v in self.values]

    def equal_to(self, other: BasePointValue) -> bool:
        if not isinstance(other, VectorValue):
            return False
        return self.value_type == other.value_type and self.values == other.values

    @staticmethod
    def create_from_numeric(
        raw_result_value: Iterable[RawResultType],
        value_type: Literal["bool", "int", "float"],
        name: str | None = None,
    ) -> VectorValue:
        val = [portablize(value_type, rv) for rv in raw_result_value]
        return VectorValue(type="vector", value_type=value_type, values=val, name=name)


type ParamType = tuple[ScalarValue, ...]
type ResultType = ScalarValue | VectorValue



from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

from lite_dist2.expections import LD2ModelTypeError, LD2UndefinedError
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace, ParameterAlignedSpaceModel
from lite_dist2.value_models.jagged_space import ParameterJaggedSpace, ParameterJaggedSpaceModel
from lite_dist2.value_models.point import ParamType, ResultType, ScalarValue, VectorValue

if TYPE_CHECKING:
    from lite_dist2.type_definitions import RawParamType, RawResultType
    from lite_dist2.value_models.base_space import FlattenSegment, ParameterSpace


class Mapping(BaseModel):
    params: ParamType
    result: ResultType


class TrialStatus(str, Enum):
    running = "running"
    done = "done"


class TrialModel(BaseModel):
    study_id: str
    trial_id: str
    timestamp: datetime
    trial_status: TrialStatus
    parameter_space: ParameterAlignedSpaceModel | ParameterJaggedSpaceModel
    result_type: Literal["scalar", "vector"]
    result_value_type: Literal["bool", "int", "float"]
    worker_node_name: str | None
    worker_node_id: str
    results: list[Mapping] | None = None


class Trial:
    def __init__(
        self,
        study_id: str,
        trial_id: str,
        timestamp: datetime,
        trial_status: TrialStatus,
        parameter_space: ParameterSpace,
        result_type: Literal["scalar", "vector"],
        result_value_type: Literal["bool", "int", "float"],
        worker_node_name: str | None,
        worker_node_id: str,
        results: list[Mapping] | None = None,
    ) -> None:
        self.study_id = study_id
        self.trial_id = trial_id
        self.timestamp = timestamp
        self.trial_status = trial_status
        self.parameter_space = parameter_space
        self.result_type = result_type
        self.result_value_type = result_value_type
        self.worker_node_name = worker_node_name
        self.worker_node_id = worker_node_id
        self.result = results

    def convert_mappings_from(self, raw_mappings: list[tuple[RawParamType, RawResultType]]) -> list[Mapping]:
        mappings = []
        for raw_param, raw_res in raw_mappings:
            param = self.parameter_space.value_tuple_to_param_type(raw_param)
            result = self._create_result_value(raw_res)
            mappings.append(Mapping(params=param, result=result))
        return mappings

    def set_result(self, mappings: list[Mapping]) -> None:
        self.result = mappings

    def get_running_segments(self) -> list[FlattenSegment]:
        if self.trial_status == TrialStatus.running:
            return self.parameter_space.get_flatten_ambient_start_and_size_list()
        return []

    def _create_result_value(self, raw_result: RawResultType) -> ResultType:
        match self.result_type:
            case "scalar":
                return ScalarValue.create_from_numeric(raw_result, self.result_value_type)
            case "vector":
                return VectorValue.create_from_numeric(raw_result, self.result_value_type)
            case _:
                raise LD2ModelTypeError(self.result_type)

    def measure_seconds_from_registered(self, now: datetime) -> int:
        delta = now - self.timestamp
        return int(delta.total_seconds())

    def find_target_value(self, target_value: ResultType) -> Mapping | None:
        # find_exact 用
        if not self.result:
            return None
        if self.trial_status != TrialStatus.done:
            return None

        for mapping in self.result:
            if mapping.result.equal_to(target_value):
                return mapping
        return None

    def to_model(self) -> TrialModel:
        return TrialModel(
            study_id=self.study_id,
            trial_id=self.trial_id,
            timestamp=self.timestamp,
            trial_status=self.trial_status,
            parameter_space=self.parameter_space.to_model(),
            result_type=self.result_type,
            result_value_type=self.result_value_type,
            worker_node_name=self.worker_node_name,
            worker_node_id=self.worker_node_id,
            results=self.result,
        )

    @staticmethod
    def from_model(model: TrialModel) -> Trial:
        match model.parameter_space.type:
            case "aligned":
                parameter_space = ParameterAlignedSpace.from_model(model.parameter_space)
            case "jagged":
                parameter_space = ParameterJaggedSpace.from_model(model.parameter_space)
            case _:
                raise LD2UndefinedError(model.parameter_space.type)
        return Trial(
            study_id=model.study_id,
            trial_id=model.trial_id,
            timestamp=model.timestamp,
            trial_status=model.trial_status,
            parameter_space=parameter_space,
            result_type=model.result_type,
            result_value_type=model.result_value_type,
            worker_node_name=model.worker_node_name,
            worker_node_id=model.worker_node_id,
            results=model.results,
        )



from collections.abc import Iterable

type PrimitiveValueType = int | float | bool
type PortableValueType = bool | str
type RawParamType = tuple[PrimitiveValueType, ...]
type RawResultType = Iterable[PrimitiveValueType] | PrimitiveValueType



import argparse
import asyncio
import logging
import socket
from pathlib import Path
from threading import Event, Thread

import uvicorn

from lite_dist2.config import TableConfigProvider
from lite_dist2.curriculum_models.curriculum import CurriculumProvider
from lite_dist2.table_node_api.api import app

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def _get_local_ip() -> str:
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            s.connect(("8.8.8.8", 80))
            return s.getsockname()[0]
    except OSError:
        logger.warning("Cannot reach DNS Server. It may show localhost IP.")
        return socket.gethostbyname(socket.gethostname())


async def _periodic_save() -> None:
    interval = TableConfigProvider.get().curriculum_save_interval_seconds
    while True:
        await asyncio.sleep(interval)
        logger.info("Performing periodic save of curriculum data")
        await CurriculumProvider.save_async()


def _run_periodic_save() -> None:
    asyncio.run(_periodic_save())


async def _periodic_timeout_check() -> None:
    interval = TableConfigProvider.get().timeout_check_interval_seconds
    while True:
        await asyncio.sleep(interval)
        logger.info("Performing periodic timeout check of trials")
        await CurriculumProvider.check_timeout()


def _run_periodic_timeout_check() -> None:
    asyncio.run(_periodic_timeout_check())


def start() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", help="Path to table node config file.", type=str, default=None)
    args = parser.parse_args()
    table_config_path = None if args.config is None else Path(args.config)

    logger.info("Table Node IP: %s", _get_local_ip())
    table_config = TableConfigProvider.get(table_config_path)

    # save thread
    save_thread = Thread(target=_run_periodic_save, daemon=True)
    save_thread.start()

    # timeout check thread
    timeout_thread = Thread(target=_run_periodic_timeout_check, daemon=True)
    timeout_thread.start()

    port = table_config.port
    uvicorn.run(app, host="0.0.0.0", port=port)  # noqa: S104


class StoppableThread(Thread):
    def __init__(self, *args: object, **kwargs: object) -> None:
        super().__init__(*args, **kwargs)
        self._stop_event = Event()

    def stop(self) -> None:
        self._stop_event.set()

    def stopped(self) -> bool:
        return self._stop_event.is_set()


def start_in_thread() -> StoppableThread:
    table_thread = StoppableThread(target=start, daemon=True)
    table_thread.start()
    return table_thread


if __name__ == "__main__":
    start()



from __future__ import annotations

from pydantic import BaseModel, Field

from lite_dist2.curriculum_models.study_portables import StudyRegistry
from lite_dist2.curriculum_models.trial import TrialModel


class BaseParam(BaseModel):
    pass


class StudyRegisterParam(BaseParam):
    study: StudyRegistry = Field(
        ...,
        description="`Study` to register",
    )


class TrialReserveParam(BaseParam):
    retaining_capacity: set[str] = Field(
        ...,
        description="List of capabilities that the worker node has.",
    )
    max_size: int = Field(
        ...,
        description="The maximum size of parameter space reserving.",
    )
    worker_node_name: str | None = Field(
        default=None,
        description="Name of the worker node. ",
    )
    worker_node_id: str = Field(
        ...,
        description="ID of the worker node",
    )


class TrialRegisterParam(BaseModel):
    trial: TrialModel = Field(
        ...,
        description="Registering trial to the table node.",
    )



from __future__ import annotations

import json
import pathlib
from typing import TYPE_CHECKING

import pytest

from lite_dist2.curriculum_models.curriculum import Curriculum, CurriculumModel
from lite_dist2.curriculum_models.study import Study
from lite_dist2.curriculum_models.study_portables import StudyModel, StudyStorage
from lite_dist2.curriculum_models.study_status import StudyStatus
from lite_dist2.curriculum_models.trial import Mapping, Trial, TrialModel, TrialStatus
from lite_dist2.curriculum_models.trial_table import TrialTable, TrialTableModel
from lite_dist2.expections import LD2ParameterError
from lite_dist2.study_strategies import BaseStudyStrategy, StudyStrategyModel
from lite_dist2.study_strategies.all_calculation_study_strategy import AllCalculationStudyStrategy
from lite_dist2.suggest_strategies import BaseSuggestStrategy, SequentialSuggestStrategy, SuggestStrategyModel
from lite_dist2.suggest_strategies.base_suggest_strategy import SuggestStrategyParam
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace
from lite_dist2.value_models.line_segment import ParameterRangeInt
from lite_dist2.value_models.point import ScalarValue
from tests.const import DT

if TYPE_CHECKING:
    from lite_dist2.value_models.base_space import ParameterSpace

_DUMMY_PARAMETER_SPACE = ParameterAlignedSpace(
    axes=[
        ParameterRangeInt(name="x", type="int", size=2, start=0, ambient_size=2, ambient_index=0),
        ParameterRangeInt(name="y", type="int", size=2, start=0, ambient_size=2, ambient_index=0),
    ],
    check_lower_filling=True,
)

_DUMMY_STUDY_STRATEGY_MODEL = StudyStrategyModel(type="all_calculation", study_strategy_param=None)
_DUMMY_SUGGEST_STRATEGY_MODEL = SuggestStrategyModel(
    type="sequential",
    suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
)
_SUMMY_TRIAL_TABLE = TrialTable(
    trials=[
        Trial(
            study_id="01",
            trial_id="01",
            timestamp=DT,
            trial_status=TrialStatus.done,
            parameter_space=_DUMMY_PARAMETER_SPACE,
            result_type="scalar",
            result_value_type="float",
            worker_node_name="w01",
            worker_node_id="w01",
            results=[
                Mapping(
                    params=(
                        ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                        ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                    ),
                    result=ScalarValue(
                        type="scalar",
                        value_type="float",
                        value="0x1.0000000000000p-1",
                    ),
                ),
                Mapping(
                    params=(
                        ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                        ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                    ),
                    result=ScalarValue(
                        type="scalar",
                        value_type="float",
                        value="0x1.0000000000000p-2",
                    ),
                ),
            ],
        ),
        Trial(
            study_id="01",
            trial_id="01",
            timestamp=DT,
            trial_status=TrialStatus.running,
            parameter_space=_DUMMY_PARAMETER_SPACE,
            result_type="scalar",
            result_value_type="float",
            worker_node_name="w01",
            worker_node_id="w01",
        ),
    ],
    aggregated_parameter_space={
        -1: [],
        0: [_DUMMY_PARAMETER_SPACE],
    },
)


class MockStudyStrategy(BaseStudyStrategy):
    def __init__(self) -> None:
        super().__init__(None)

    def extract_mappings(self, trial_table: TrialTable) -> list[Mapping]:
        pass

    def is_done(self, trial_table: TrialTable, parameter_space: ParameterAlignedSpace) -> bool:
        pass

    def to_model(self) -> StudyStrategyModel:
        return _DUMMY_STUDY_STRATEGY_MODEL


class MockSuggestStrategy(BaseSuggestStrategy):
    def __init__(self) -> None:
        super().__init__(SuggestStrategyParam(strict_aligned=False), _DUMMY_PARAMETER_SPACE)

    def suggest(self, trial_table: TrialTable, max_num: int) -> ParameterSpace:
        pass

    def to_model(self) -> SuggestStrategyModel:
        return _DUMMY_SUGGEST_STRATEGY_MODEL


class MockStudy(Study):
    def __init__(self, study_id: str, done: bool) -> None:
        super().__init__(
            study_id,
            study_id,
            set(),
            StudyStatus.wait,
            DT,
            MockStudyStrategy(),
            MockSuggestStrategy(),
            _DUMMY_PARAMETER_SPACE,
            "scalar",
            "int",
            TrialTable([], None),
        )
        self.study_id = study_id
        self._done = done

    def is_done(self) -> bool:
        return self._done

    def update_status(self) -> None:
        pass

    def to_storage(self) -> StudyStorage:
        return StudyStorage(
            study_id=self.study_id,
            name=self.study_id,
            required_capacity=set(),
            registered_timestamp=DT,
            study_strategy=self.study_strategy.to_model(),
            suggest_strategy=self.suggest_strategy.to_model(),
            parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
            done_timestamp=DT,
            result_type="scalar",
            result_value_type="int",
            results=[],
            done_grids=4,
        )


@pytest.fixture
def done_study_fixture() -> MockStudy:
    return MockStudy(study_id="done_study", done=True)


@pytest.fixture
def not_done_study_fixture() -> MockStudy:
    return MockStudy(study_id="not_done_study", done=False)


def test_curriculum_to_storage_if_done(done_study_fixture: MockStudy, not_done_study_fixture: MockStudy) -> None:
    curriculum = Curriculum(studies=[done_study_fixture, not_done_study_fixture], storages=[])
    curriculum.to_storage_if_done()

    assert len(curriculum.studies) == 1
    assert len(curriculum.storages) == 1
    assert curriculum.studies[0].study_id == "not_done_study"
    assert curriculum.storages[0].study_id == "done_study"


@pytest.fixture
def sample_curriculum_fixture() -> Curriculum:
    studies = [
        Study.from_model(
            StudyModel(
                study_id="01",
                name="s1",
                required_capacity=set(),
                status=StudyStatus.running,
                registered_timestamp=DT,
                study_strategy=StudyStrategyModel(type="all_calculation", study_strategy_param=None),
                suggest_strategy=SuggestStrategyModel(
                    type="sequential",
                    suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
                ),
                parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
                result_type="scalar",
                result_value_type="int",
                trial_table=TrialTableModel(
                    trials=[
                        TrialModel(
                            study_id="01",
                            trial_id="01",
                            timestamp=DT,
                            trial_status=TrialStatus.done,
                            parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
                            result_type="scalar",
                            result_value_type="float",
                            worker_node_name="w01",
                            worker_node_id="w01",
                            results=[
                                Mapping(
                                    params=(
                                        ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                        ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                    ),
                                    result=ScalarValue(
                                        type="scalar",
                                        value_type="float",
                                        value="0x1.0000000000000p-1",
                                    ),
                                ),
                                Mapping(
                                    params=(
                                        ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                        ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                    ),
                                    result=ScalarValue(
                                        type="scalar",
                                        value_type="float",
                                        value="0x1.0000000000000p-2",
                                    ),
                                ),
                            ],
                        ),
                        TrialModel(
                            study_id="01",
                            trial_id="01",
                            timestamp=DT,
                            trial_status=TrialStatus.running,
                            parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
                            result_type="scalar",
                            result_value_type="float",
                            worker_node_name="w01",
                            worker_node_id="w01",
                        ),
                    ],
                    aggregated_parameter_space={
                        -1: [],
                        0: [_DUMMY_PARAMETER_SPACE.to_model()],
                    },
                ),
            ),
        ),
    ]
    storages = [
        StudyStorage(
            study_id="s2",
            name="s2",
            required_capacity=set(),
            registered_timestamp=DT,
            study_strategy=_DUMMY_STUDY_STRATEGY_MODEL,
            suggest_strategy=_DUMMY_SUGGEST_STRATEGY_MODEL,
            parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
            done_timestamp=DT,
            result_type="scalar",
            result_value_type="int",
            results=[],
            done_grids=4,
        ),
    ]
    return Curriculum(studies, storages)


def test_curriculum_save_and_load(tmp_path: str, sample_curriculum_fixture: Curriculum) -> None:
    json_path = pathlib.Path(f"{tmp_path}/curriculum.json")
    assert not json_path.exists()

    sample_curriculum_fixture.save(json_path)
    assert json_path.exists()

    with json_path.open("r", encoding="utf-8") as f:
        json_data = json.load(f)
    model = CurriculumModel.model_validate(json_data)
    assert model is not None

    loaded_curriculum = Curriculum.load_or_create(json_path)
    assert len(loaded_curriculum.studies) == len(sample_curriculum_fixture.studies)
    assert len(loaded_curriculum.storages) == len(sample_curriculum_fixture.storages)
    assert loaded_curriculum.studies[0].name == sample_curriculum_fixture.studies[0].name
    assert loaded_curriculum.storages[0].name == sample_curriculum_fixture.storages[0].name


def test_curriculum_load_or_create_empty(tmp_path: str) -> None:
    json_path = pathlib.Path(f"{tmp_path}/non_existent.json")
    curriculum = Curriculum.load_or_create(json_path)

    assert isinstance(curriculum, Curriculum)
    assert len(curriculum.studies) == 0
    assert len(curriculum.storages) == 0


@pytest.mark.parametrize(
    ("retaining_capacity", "expected_study_id"),
    [
        pytest.param(
            {"hash", "preimage"},
            "hash_2",
            id="obtain running",
        ),
        pytest.param(
            {"hash"},
            "hash_1",
            id="obtain low required wait",
        ),
        pytest.param(
            {"mandelbrot"},
            None,
            id="obtain nothing",
        ),
    ],
)
def test_curriculum_get_available_study(retaining_capacity: set[str], expected_study_id: str | None) -> None:
    _study_param = {
        "name": "",
        "registered_timestamp": DT,
        "study_strategy": AllCalculationStudyStrategy(None),
        "suggest_strategy": SequentialSuggestStrategy(
            SuggestStrategyParam(strict_aligned=True),
            _DUMMY_PARAMETER_SPACE,
        ),
        "parameter_space": _DUMMY_PARAMETER_SPACE,
        "result_type": "scalar",
        "result_value_type": "int",
        "trial_table": TrialTable.from_model(TrialTableModel.create_empty()),
    }
    curriculum = Curriculum(
        studies=[
            Study(
                study_id="hash_1",
                required_capacity={"hash"},
                status=StudyStatus.wait,
                **_study_param,
            ),
            Study(
                study_id="hash_2",
                required_capacity={"hash", "preimage"},
                status=StudyStatus.running,
                **_study_param,
            ),
            Study(
                study_id="hash_3",
                required_capacity={"hash", "preimage"},
                status=StudyStatus.wait,
                **_study_param,
            ),
        ],
        storages=[],
    )

    study = curriculum.get_available_study(retaining_capacity)
    assert (study.study_id if study is not None else None) == expected_study_id


@pytest.mark.parametrize(
    ("study_id", "name", "expected_id", "expected_storages"),
    [
        pytest.param(
            "xxx",
            "xxx",
            None,
            [
                StudyStorage(
                    study_id="s01",
                    name="n01",
                    required_capacity=set(),
                    registered_timestamp=DT,
                    study_strategy=_DUMMY_STUDY_STRATEGY_MODEL,
                    suggest_strategy=_DUMMY_SUGGEST_STRATEGY_MODEL,
                    parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
                    done_timestamp=DT,
                    result_type="scalar",
                    result_value_type="int",
                    results=[],
                    done_grids=4,
                ),
                StudyStorage(
                    study_id="s02",
                    name="n02",
                    required_capacity=set(),
                    registered_timestamp=DT,
                    study_strategy=_DUMMY_STUDY_STRATEGY_MODEL,
                    suggest_strategy=_DUMMY_SUGGEST_STRATEGY_MODEL,
                    parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
                    done_timestamp=DT,
                    result_type="scalar",
                    result_value_type="int",
                    results=[],
                    done_grids=4,
                ),
            ],
            id="None",
        ),
        pytest.param(
            "s01",
            None,
            "s01",
            [
                StudyStorage(
                    study_id="s02",
                    name="n02",
                    required_capacity=set(),
                    registered_timestamp=DT,
                    study_strategy=_DUMMY_STUDY_STRATEGY_MODEL,
                    suggest_strategy=_DUMMY_SUGGEST_STRATEGY_MODEL,
                    parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
                    done_timestamp=DT,
                    result_type="scalar",
                    result_value_type="int",
                    results=[],
                    done_grids=4,
                ),
            ],
            id="id pickup",
        ),
        pytest.param(
            None,
            "n02",
            "s02",
            [
                StudyStorage(
                    study_id="s01",
                    name="n01",
                    required_capacity=set(),
                    registered_timestamp=DT,
                    study_strategy=_DUMMY_STUDY_STRATEGY_MODEL,
                    suggest_strategy=_DUMMY_SUGGEST_STRATEGY_MODEL,
                    parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
                    done_timestamp=DT,
                    result_type="scalar",
                    result_value_type="int",
                    results=[],
                    done_grids=4,
                ),
            ],
            id="name pickup",
        ),
        pytest.param(
            "s01",
            "n02",
            "s01",
            [
                StudyStorage(
                    study_id="s02",
                    name="n02",
                    required_capacity=set(),
                    registered_timestamp=DT,
                    study_strategy=_DUMMY_STUDY_STRATEGY_MODEL,
                    suggest_strategy=_DUMMY_SUGGEST_STRATEGY_MODEL,
                    parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
                    done_timestamp=DT,
                    result_type="scalar",
                    result_value_type="int",
                    results=[],
                    done_grids=4,
                ),
            ],
            id="prior id",
        ),
    ],
)
def test_curriculum_pop_storage(
    study_id: str | None,
    name: str | None,
    expected_id: str | None,
    expected_storages: list[StudyStorage],
) -> StudyStorage | None:
    curr = Curriculum(
        studies=[],
        storages=[
            StudyStorage(
                study_id="s01",
                name="n01",
                required_capacity=set(),
                registered_timestamp=DT,
                study_strategy=_DUMMY_STUDY_STRATEGY_MODEL,
                suggest_strategy=_DUMMY_SUGGEST_STRATEGY_MODEL,
                parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
                done_timestamp=DT,
                result_type="scalar",
                result_value_type="int",
                results=[],
                done_grids=4,
            ),
            StudyStorage(
                study_id="s02",
                name="n02",
                required_capacity=set(),
                registered_timestamp=DT,
                study_strategy=_DUMMY_STUDY_STRATEGY_MODEL,
                suggest_strategy=_DUMMY_SUGGEST_STRATEGY_MODEL,
                parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
                done_timestamp=DT,
                result_type="scalar",
                result_value_type="int",
                results=[],
                done_grids=4,
            ),
        ],
    )

    popped = curr.pop_storage(study_id, name)
    if expected_id is None:
        assert popped is None
    else:
        assert popped.study_id == expected_id
    assert curr.storages == expected_storages


def test_curriculum_pop_storage_raises() -> None:
    curr = Curriculum(studies=[], storages=[])
    with pytest.raises(LD2ParameterError):
        _ = curr.pop_storage(None, None)


@pytest.mark.parametrize(
    ("study_id", "name", "expected_cancel_result", "expected_studies"),
    [
        pytest.param(
            "s01",
            None,
            True,
            [
                Study(
                    study_id="s02",
                    name="n02",
                    required_capacity=set(),
                    status=StudyStatus.running,
                    registered_timestamp=DT,
                    study_strategy=AllCalculationStudyStrategy(_DUMMY_STUDY_STRATEGY_MODEL.study_strategy_param),
                    suggest_strategy=SequentialSuggestStrategy(
                        _DUMMY_SUGGEST_STRATEGY_MODEL.suggest_strategy_param,
                        _DUMMY_PARAMETER_SPACE,
                    ),
                    parameter_space=_DUMMY_PARAMETER_SPACE,
                    result_type="scalar",
                    result_value_type="int",
                    trial_table=_SUMMY_TRIAL_TABLE,
                ),
            ],
            id="cancel by id",
        ),
        pytest.param(
            None,
            "n02",
            True,
            [
                Study(
                    study_id="s01",
                    name="n01",
                    required_capacity=set(),
                    status=StudyStatus.wait,
                    registered_timestamp=DT,
                    study_strategy=AllCalculationStudyStrategy(_DUMMY_STUDY_STRATEGY_MODEL.study_strategy_param),
                    suggest_strategy=SequentialSuggestStrategy(
                        _DUMMY_SUGGEST_STRATEGY_MODEL.suggest_strategy_param,
                        _DUMMY_PARAMETER_SPACE,
                    ),
                    parameter_space=_DUMMY_PARAMETER_SPACE,
                    result_type="scalar",
                    result_value_type="int",
                    trial_table=_SUMMY_TRIAL_TABLE,
                ),
            ],
            id="cancel by name",
        ),
        pytest.param(
            "xxxx",
            None,
            False,
            [
                Study(
                    study_id="s01",
                    name="n01",
                    required_capacity=set(),
                    status=StudyStatus.wait,
                    registered_timestamp=DT,
                    study_strategy=AllCalculationStudyStrategy(_DUMMY_STUDY_STRATEGY_MODEL.study_strategy_param),
                    suggest_strategy=SequentialSuggestStrategy(
                        _DUMMY_SUGGEST_STRATEGY_MODEL.suggest_strategy_param,
                        _DUMMY_PARAMETER_SPACE,
                    ),
                    parameter_space=_DUMMY_PARAMETER_SPACE,
                    result_type="scalar",
                    result_value_type="int",
                    trial_table=_SUMMY_TRIAL_TABLE,
                ),
                Study(
                    study_id="s02",
                    name="n02",
                    required_capacity=set(),
                    status=StudyStatus.running,
                    registered_timestamp=DT,
                    study_strategy=AllCalculationStudyStrategy(_DUMMY_STUDY_STRATEGY_MODEL.study_strategy_param),
                    suggest_strategy=SequentialSuggestStrategy(
                        _DUMMY_SUGGEST_STRATEGY_MODEL.suggest_strategy_param,
                        _DUMMY_PARAMETER_SPACE,
                    ),
                    parameter_space=_DUMMY_PARAMETER_SPACE,
                    result_type="scalar",
                    result_value_type="int",
                    trial_table=_SUMMY_TRIAL_TABLE,
                ),
            ],
            id="Not found",
        ),
    ],
)
def test_curriculum_cancel_study(
    study_id: str | None,
    name: str | None,
    expected_cancel_result: bool,
    expected_studies: list[Study],
) -> None:
    curr = Curriculum(
        storages=[],
        studies=[
            Study(
                study_id="s01",
                name="n01",
                required_capacity=set(),
                status=StudyStatus.wait,
                registered_timestamp=DT,
                study_strategy=AllCalculationStudyStrategy(_DUMMY_STUDY_STRATEGY_MODEL.study_strategy_param),
                suggest_strategy=SequentialSuggestStrategy(
                    _DUMMY_SUGGEST_STRATEGY_MODEL.suggest_strategy_param,
                    _DUMMY_PARAMETER_SPACE,
                ),
                parameter_space=_DUMMY_PARAMETER_SPACE,
                result_type="scalar",
                result_value_type="int",
                trial_table=_SUMMY_TRIAL_TABLE,
            ),
            Study(
                study_id="s02",
                name="n02",
                required_capacity=set(),
                status=StudyStatus.running,
                registered_timestamp=DT,
                study_strategy=AllCalculationStudyStrategy(_DUMMY_STUDY_STRATEGY_MODEL.study_strategy_param),
                suggest_strategy=SequentialSuggestStrategy(
                    _DUMMY_SUGGEST_STRATEGY_MODEL.suggest_strategy_param,
                    _DUMMY_PARAMETER_SPACE,
                ),
                parameter_space=_DUMMY_PARAMETER_SPACE,
                result_type="scalar",
                result_value_type="int",
                trial_table=_SUMMY_TRIAL_TABLE,
            ),
        ],
    )

    actual_cancel_result = curr.cancel_study(study_id, name)
    assert actual_cancel_result == expected_cancel_result

    assert len(curr.studies) == len(expected_studies)
    for actual_study, expected_study in zip(curr.studies, expected_studies, strict=True):
        assert actual_study.to_model() == expected_study.to_model()


def test_curriculum_cancel_study_raises() -> None:
    curr = Curriculum(studies=[], storages=[])
    with pytest.raises(LD2ParameterError):
        _ = curr.cancel_study(None, None)


@pytest.mark.parametrize(
    ("study_id", "name", "expected"),
    [
        pytest.param("xxx", "xxx", StudyStatus.not_found, id="not_found"),
        pytest.param("s01", None, StudyStatus.done, id="done (storage) by id"),
        pytest.param(None, "n01", StudyStatus.done, id="done (storage) by name"),
        pytest.param("s04", None, StudyStatus.done, id="done (study) by id"),
        pytest.param(None, "n02", StudyStatus.running, id="running by name"),
        pytest.param("s03", None, StudyStatus.wait, id="wait by id"),
        pytest.param("s03", "n02", StudyStatus.wait, id="prior id"),
    ],
)
def test_curriculum_get_study_status(study_id: str | None, name: str | None, expected: StudyStatus) -> None:
    _study_param = {
        "required_capacity": {"hash"},
        "registered_timestamp": DT,
        "study_strategy": AllCalculationStudyStrategy(None),
        "suggest_strategy": SequentialSuggestStrategy(
            SuggestStrategyParam(strict_aligned=True),
            _DUMMY_PARAMETER_SPACE,
        ),
        "parameter_space": _DUMMY_PARAMETER_SPACE,
        "result_type": "scalar",
        "result_value_type": "int",
        "trial_table": TrialTable.from_model(TrialTableModel.create_empty()),
    }
    curr = Curriculum(
        studies=[
            Study(
                study_id="s02",
                name="n02",
                status=StudyStatus.running,
                **_study_param,
            ),
            Study(
                study_id="s03",
                name="n03",
                status=StudyStatus.wait,
                **_study_param,
            ),
            Study(
                study_id="s04",
                name="n04",
                status=StudyStatus.done,
                **_study_param,
            ),
        ],
        storages=[
            StudyStorage(
                study_id="s01",
                name="n01",
                required_capacity=set(),
                registered_timestamp=DT,
                study_strategy=_DUMMY_STUDY_STRATEGY_MODEL,
                suggest_strategy=_DUMMY_SUGGEST_STRATEGY_MODEL,
                parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
                done_timestamp=DT,
                result_type="scalar",
                result_value_type="int",
                results=[],
                done_grids=4,
            ),
        ],
    )

    actual = curr.get_study_status(study_id, name)
    assert actual == expected


def test_curriculum_get_study_status_raises() -> None:
    curr = Curriculum(studies=[], storages=[])
    with pytest.raises(LD2ParameterError):
        _ = curr.get_study_status(None, None)


@pytest.mark.parametrize(
    ("name", "expected"),
    [
        pytest.param("n01", False, id="storage: False"),
        pytest.param("n02", False, id="running: False"),
        pytest.param("n03", False, id="wait: False"),
        pytest.param("n04", False, id="done: False"),
        pytest.param("n07", True, id="unique name: True"),
        pytest.param(None, True, id="None: True"),
    ],
)
def test_curriculum_try_insert_study(name: str | None, expected: bool) -> None:
    _study_param = {
        "required_capacity": {"hash"},
        "registered_timestamp": DT,
        "study_strategy": AllCalculationStudyStrategy(None),
        "suggest_strategy": SequentialSuggestStrategy(
            SuggestStrategyParam(strict_aligned=True),
            _DUMMY_PARAMETER_SPACE,
        ),
        "parameter_space": _DUMMY_PARAMETER_SPACE,
        "result_type": "scalar",
        "result_value_type": "int",
        "trial_table": TrialTable.from_model(TrialTableModel.create_empty()),
    }
    curr = Curriculum(
        studies=[
            Study(
                study_id="s02",
                name="n02",
                status=StudyStatus.running,
                **_study_param,
            ),
            Study(
                study_id="s03",
                name="n03",
                status=StudyStatus.wait,
                **_study_param,
            ),
            Study(
                study_id="s04",
                name="n04",
                status=StudyStatus.done,
                **_study_param,
            ),
            Study(
                study_id="s05",
                name=None,
                status=StudyStatus.done,
                **_study_param,
            ),
        ],
        storages=[
            StudyStorage(
                study_id="s01",
                name="n01",
                required_capacity=set(),
                registered_timestamp=DT,
                study_strategy=_DUMMY_STUDY_STRATEGY_MODEL,
                suggest_strategy=_DUMMY_SUGGEST_STRATEGY_MODEL,
                parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
                done_timestamp=DT,
                result_type="scalar",
                result_value_type="int",
                results=[],
                done_grids=4,
            ),
            StudyStorage(
                study_id="s06",
                name=None,
                required_capacity=set(),
                registered_timestamp=DT,
                study_strategy=_DUMMY_STUDY_STRATEGY_MODEL,
                suggest_strategy=_DUMMY_SUGGEST_STRATEGY_MODEL,
                parameter_space=_DUMMY_PARAMETER_SPACE.to_model(),
                done_timestamp=DT,
                result_type="scalar",
                result_value_type="int",
                results=[],
                done_grids=4,
            ),
        ],
    )
    new_study = Study(
        name=name,
        study_id="s07",
        status=StudyStatus.running,
        **_study_param,
    )
    actual = curr.try_insert_study(new_study)
    assert actual == expected
